"""The job configuration class for collecting all jobs."""

import os
from jade.jobs.job_inputs_interface import JobInputsInterface
from jade.extensions.generic_command import GenericCommandParameters


class GenericCommandInputs(JobInputsInterface):
    """A class used for configuring generic_command jobs."""

    def __init__(self, filename):
        """
        Parameters
        ----------
        filename : str
            Input file containing commands, one line per command

        """
        self._filename = filename
        self._parameters = []
        with open(filename) as f_in:
            self._parameters = []
            for line in f_in.readlines():
                line = line.strip()
                if line:
                    self._parameters.append(GenericCommandParameters(command=line))

    @property
    def base_directory(self):
        return os.path.dirname(self._filename)

    def get_available_parameters(self):
        return self._parameters

    def iter_jobs(self):
        """Return a list of auto-regression jobs"""
        for param in self._parameters:
            yield param



"""Manages multi-node job coordination"""

import logging
import os
import socket
import time
from pathlib import Path

import click

from jade.loggers import setup_logging
from jade.utils.run_command import check_run_command, run_command
from jade.utils.utils import get_cli_string


SHUTDOWN_FILENAME = "shutdown"

logger = logging.getLogger(__name__)


@click.command(
    context_settings=dict(
        allow_extra_args=True,
        ignore_unknown_options=True,
    )
)
@click.argument("job_name")
@click.option(
    "-o",
    "--jade-runtime-output",
    required=True,
    type=click.Path(exists=True),
    help="output directory passed to 'jade submit-jobs'",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
@click.argument("manager_script_and_args", nargs=-1, type=click.UNPROCESSED)
def run_multi_node_job(job_name, jade_runtime_output, verbose, manager_script_and_args):
    """Run a job across multiple nodes. The manager node will invoke manager_script_and_args."""
    output = {}
    check_run_command(f"jade cluster am-i-manager {jade_runtime_output}", output)
    result = output["stdout"].strip()
    if result == "true":
        ret = run_manager(job_name, jade_runtime_output, verbose, manager_script_and_args)
    else:
        assert result == "false", result
        # The only purpose of this worker function is to keep the node allocation
        # alive. There are more efficient ways of doing this with HPC commands.
        # However, this procedure allows us to run the JADE JobRunner in the
        # background on each node and collect resource utilization statistics.
        ret = run_worker(job_name, jade_runtime_output, verbose)

    return ret


def run_manager(job_name, output_dir, verbose, manager_script_and_args):
    """Run the manager instance."""
    shutdown_file = _get_shutdown_file(job_name, output_dir)
    if shutdown_file.exists():
        os.remove(shutdown_file)
    try:
        return _run_manager(job_name, output_dir, verbose, manager_script_and_args)
    finally:
        # Notify the workers to shutdown.
        shutdown_file.touch()


def _run_manager(job_name, output_dir, verbose, manager_script_and_args):
    filename = os.path.join(output_dir, f"run_multi_node_job_manager__{job_name}.log")
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, filename, file_level=level, console_level=level, mode="w")
    logger.info("Run manager on %s: %s", socket.gethostname(), get_cli_string())

    # Note that the manager receives its own hostname.
    output = {}
    job_id = os.environ["SLURM_JOB_ID"]  # TODO: needs to be agnostic to HPC type
    check_run_command(f"jade cluster hostnames -j {job_id} {output_dir}", output)
    hostnames = [x for x in output["stdout"].split() if x != ""]
    logger.info("Manager found %s hostnames: %s", len(hostnames), hostnames)
    cmd = " ".join(manager_script_and_args)
    logger.info("Run manager script [%s]", cmd)

    os.environ["JADE_OUTPUT_DIR"] = output_dir
    os.environ["JADE_COMPUTE_NODE_NAMES"] = " ".join(hostnames)
    start = time.time()
    ret = run_command(cmd)
    logger.info("Finished job. duration = %s seconds", time.time() - start)
    return ret


def run_worker(job_name, output_dir, verbose, poll_interval=60):
    """Run a worker instance."""
    hostname = socket.gethostname()
    filename = os.path.join(output_dir, f"run_multi_node_job_worker__{job_name}__{hostname}.log")
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, filename, file_level=level, console_level=level, mode="w")
    logger.info("Run worker: %s", get_cli_string())

    shutdown_file = _get_shutdown_file(job_name, output_dir)
    while not shutdown_file.exists():
        logger.debug("sleep for %s seconds", poll_interval)
        time.sleep(poll_interval)

    logger.info("Detected shutdown.")
    return 0


def _get_shutdown_file(job_name, output_dir):
    return Path(output_dir) / (SHUTDOWN_FILENAME + "__" + job_name)



"""Contains base class for simulation configurations."""

import abc
import enum
import json
import logging
import os
import sys
from collections import defaultdict
from datetime import timedelta

import toml

from jade.common import CONFIG_FILE
from jade.exceptions import InvalidConfiguration, InvalidParameter
from jade.extensions.registry import Registry, ExtensionClassType
from jade.jobs.job_container_by_name import JobContainerByName
from jade.models import submission_group
from jade.models.submission_group import SubmissionGroup, SubmitterParams
from jade.utils.utils import dump_data, load_data, ExtendedJSONEncoder
from jade.utils.timing_utils import timed_debug


logger = logging.getLogger(__name__)


class ConfigSerializeOptions(enum.Enum):
    """Defines option for JobConfiguration serialization."""

    JOBS = enum.auto()
    JOB_NAMES = enum.auto()
    NO_JOB_INFO = enum.auto()


class JobConfiguration(abc.ABC):
    """Base class for any simulation configuration."""

    FILENAME_DELIMITER = "_"
    FORMAT_VERSION = "v0.2.0"

    def __init__(
        self,
        container=None,
        job_global_config=None,
        job_post_process_config=None,
        user_data=None,
        submission_groups=None,
        setup_command=None,
        teardown_command=None,
        node_setup_command=None,
        node_teardown_command=None,
        **kwargs,
    ):
        """
        Constructs JobConfiguration.

        Parameters
        ----------
        inputs : JobInputsInterface
        container : JobContainerInterface

        """
        self._jobs = container or JobContainerByName()
        self._job_names = None
        self._jobs_directory = kwargs.get("jobs_directory")
        self._registry = Registry()
        self._job_global_config = job_global_config
        self._job_post_process_config = job_post_process_config
        self._user_data = user_data or {}
        self._submission_groups = [SubmissionGroup(**x) for x in submission_groups or []]
        self._setup_command = setup_command
        self._teardown_command = teardown_command
        self._node_setup_command = node_setup_command
        self._node_teardown_command = node_teardown_command

        if kwargs.get("do_not_deserialize_jobs", False):
            assert "job_names" in kwargs, str(kwargs)
            self._job_names = kwargs["job_names"]
            return

        if "jobs" in kwargs:
            self._deserialize_jobs(kwargs["jobs"])
        elif "job_names" in kwargs:
            assert self._jobs_directory is not None, str(kwargs)
            names = kwargs["job_names"]
            self._deserialize_jobs_from_names(names)

    def __repr__(self):
        """Concisely display all instance information."""
        return self.dumps()

    def _deserialize_jobs(self, jobs):
        for _job in jobs:
            param_class = self.job_parameters_class(_job["extension"])
            job = param_class.deserialize(_job)
            self.add_job(job)

    def _deserialize_jobs_from_names(self, job_names):
        for name in job_names:
            job = self._get_job_by_name(name)
            self.add_job(job)

    def _dump(self, stream=sys.stdout, fmt=".json", indent=2):
        # Note: the default is JSON here because parsing 100 MB .toml files
        # is an order of magnitude slower.
        data = self.serialize()
        if fmt == ".json":
            json.dump(data, stream, indent=indent, cls=ExtendedJSONEncoder)
        elif fmt == ".toml":
            toml.dump(data, stream)
        else:
            assert False, fmt

    def _get_job_by_name(self, name):
        assert self._jobs_directory is not None
        filename = os.path.join(self._jobs_directory, name) + ".json"
        assert os.path.exists(filename), filename
        job = load_data(filename)
        param_class = self.job_parameters_class(job["extension"])
        return param_class.deserialize(job)

    @abc.abstractmethod
    def _serialize(self, data):
        """Create implementation-specific data for serialization."""

    def add_user_data(self, key, data):
        """Add user data referenced by a key. Must be JSON-serializable

        Parameters
        ----------
        key : str
        data : any

        Raises
        ------
        InvalidParameter
            Raised if the key is already stored.

        """
        if key in self._user_data:
            raise InvalidParameter(f"{key} is already stored. Call remove_user_data first")

        self._user_data[key] = data

    def get_user_data(self, key):
        """Get the user data associated with key.

        Parameters
        ----------
        key : str

        Returns
        -------
        any

        """
        data = self._user_data.get(key)
        if data is None:
            raise InvalidParameter(f"{key} is not stored.")

        return data

    def remove_user_data(self, key):
        """Remove the key from the user data config.

        Parameters
        ----------
        key : str

        """
        self._user_data.pop(key, None)

    def list_user_data_keys(self):
        """List the stored user data keys.

        Returns
        -------
        list
            list of str

        """
        return sorted(list(self._user_data.keys()))

    def check_job_dependencies(self):
        """Check for impossible conditions with job dependencies.

        Raises
        ------
        InvalidConfiguration
            Raised if job dependencies have an impossible condition.

        """
        # This currently only checks that all jobs defined as blocking exist.
        # It does not look for deadlocks.

        job_names = set()
        blocking_jobs = set()
        for job in self.iter_jobs():
            job_names.add(job.name)
            blocking_jobs.update(job.get_blocking_jobs())

        missing_jobs = blocking_jobs.difference(job_names)
        if missing_jobs:
            for job in missing_jobs:
                logger.error("%s is blocking a job but does not exist", job)
            raise InvalidConfiguration("job ordering definitions are invalid")

    def check_job_estimated_run_minutes(self, group_name):
        """Check that estimated_run_minutes is set for all jobs in a group."""
        missing_estimate = []
        for job in self.iter_jobs():
            if job.submission_group == group_name and job.estimated_run_minutes is None:
                missing_estimate.append(job.name)

        if missing_estimate:
            for job in missing_estimate:
                logger.error("Job %s does not define estimated_run_minutes", job)
            raise InvalidConfiguration(
                "Submitting batches by time requires that each job define estimated_run_minutes"
            )

    def check_job_runtimes(self):
        """Check for any job with a longer estimated runtime than the walltime.

        Raises
        ------
        InvalidConfiguration
            Raised if any job is too long.

        """
        wall_times = {x.name: x.submitter_params.get_wall_time() for x in self.submission_groups}
        for job in self.iter_jobs():
            wall_time = wall_times[job.submission_group]
            if job.estimated_run_minutes is not None:
                estimate = timedelta(minutes=job.estimated_run_minutes)
                if estimate > wall_time:
                    raise InvalidConfiguration(
                        f"job {job.name} has estimated_run_minutes={estimate} longer than wall_time={wall_time}"
                    )

    def check_spark_config(self):
        """If Spark jobs are present in the config, configure the params to run
        one job at a time.

        """
        groups_with_spark_jobs = set()
        for job in self.iter_jobs():
            if job.is_spark_job():
                groups_with_spark_jobs.add(job.submission_group)

        for group_name in groups_with_spark_jobs:
            for group in self._submission_groups:
                if (
                    group.name == group_name
                    and group.submitter_params.num_parallel_processes_per_node != 1
                ):
                    group.submitter_params.num_parallel_processes_per_node = 1
                    logger.info(
                        "Set num_parallel_processes_per_node=1 for group=%s for Spark jobs.",
                        group_name,
                    )

    def check_submission_groups(self):
        """Check for invalid job submission group assignments.
        Make a default group if none are defined and assign it to each job.

        Raises
        ------
        InvalidConfiguration
            Raised if submission group assignments are invalid.

        """
        first_group = next(iter(self.submission_groups))
        group_params = (
            "try_add_blocked_jobs",
            "time_based_batching",
            "num_parallel_processes_per_node",
            "hpc_config",
            "per_node_batch_size",
            "singularity_params",
            "distributed_submitter",
            "resource_monitor_stats",
        )
        user_overrides = (
            "distributed_submitter",
            "generate_reports",
            "resource_monitor_interval",
            "resource_monitor_type",
            "dry_run",
            "verbose",
        )
        user_override_if_not_set = ("node_setup_script", "node_shutdown_script")
        must_be_same = ("max_nodes", "poll_interval")
        all_params = (must_be_same, group_params, user_overrides, user_override_if_not_set)
        fields = {item for params in all_params for item in params}
        assert sorted(list(fields)) == sorted(SubmitterParams.__fields__), sorted(list(fields))
        hpc_type = first_group.submitter_params.hpc_config.hpc_type
        group_names = set()
        for group in self.submission_groups:
            if group.name in group_names:
                raise InvalidConfiguration(f"submission group {group.name} is listed twice")
            group_names.add(group.name)
            if group.submitter_params.hpc_config.hpc_type != hpc_type:
                raise InvalidConfiguration(f"hpc_type values must be the same in all groups")
            for param in must_be_same:
                first_val = getattr(first_group.submitter_params, param)
                this_val = getattr(group.submitter_params, param)
                if this_val != first_val:
                    raise InvalidConfiguration(f"{param} must be the same in all groups")
            for param in user_overrides:
                user_val = getattr(group.submitter_params, param)
                setattr(group.submitter_params, param, user_val)
            for param in user_override_if_not_set:
                user_val = getattr(group.submitter_params, param)
                group_val = getattr(group.submitter_params, param)
                if group_val is None:
                    setattr(group.submitter_params, param, user_val)

        jobs_by_group = defaultdict(list)
        for job in self.iter_jobs():
            if job.submission_group is None:
                raise InvalidConfiguration(
                    f"Job {job.name} does not have a submission group assigned"
                )
            if job.submission_group not in group_names:
                raise InvalidConfiguration(
                    f"Job {job.name} has an invalid submission group: {job.submission_group}"
                )
            jobs_by_group[job.submission_group].append(job.name)

        group_counts = {}
        for name, jobs in jobs_by_group.items():
            if not jobs:
                logger.warning("Submission group %s does not have any jobs defined", name)
            group_counts[name] = len(jobs)

        for name, count in sorted(group_counts.items()):
            logger.info("Submission group %s has %s jobs", name, count)

    def assign_default_submission_group(self, submitter_params):
        default_name = "default"
        group = SubmissionGroup(name=default_name, submitter_params=submitter_params)
        for job in self.iter_jobs():
            job.submission_group = group.name
        self.append_submission_group(group)

    @abc.abstractmethod
    def create_from_result(self, job, output_dir):
        """Create an instance from a result file.

        Parameters
        ----------
        job : JobParametersInterface
        output_dir : str

        Returns
        -------
        class

        """

    def add_job(self, job):
        """Add a job to the configuration.

        Parameters
        ----------
        job : JobParametersInterface

        """
        self._jobs.add_job(job)

    def clear(self):
        """Clear all configured jobs."""
        self._jobs.clear()

    @timed_debug
    def dump(self, filename=None, stream=sys.stdout, indent=2):
        """Convert the configuration to structured text format.

        Parameters
        ----------
        filename : str | None
            Write configuration to this file (must be .json or .toml).
            If None, write the text to stream.
            Recommend using .json for large files. .toml is much slower.
        stream : file
            File-like interface that supports write().
        indent : int
            If JSON, use this indentation.

        Raises
        ------
        InvalidParameter
            Raised if filename does not have a supported extenstion.

        """
        if filename is None and stream is None:
            raise InvalidParameter("must set either filename or stream")

        if filename is not None:
            ext = os.path.splitext(filename)[1]
            if ext not in (".json", ".toml"):
                raise InvalidParameter("Only .json and .toml are supported")

            with open(filename, "w") as f_out:
                self._dump(f_out, fmt=ext, indent=indent)
        else:
            self._dump(stream, indent=indent)

        logger.info("Dumped configuration to %s", filename)

    def dumps(self, fmt_module=toml, **kwargs):
        """Dump the configuration to a formatted string."""
        return fmt_module.dumps(self.serialize(), **kwargs)

    @classmethod
    def deserialize(cls, filename_or_data, do_not_deserialize_jobs=False):
        """Create a class instance from a saved configuration file.

        Parameters
        ----------
        filename : str | dict
            path to configuration file or that file loaded as a dict
        do_not_deserialize_jobs : bool
            Set to True to avoid the overhead of loading all jobs from disk.
            Job_names will be stored instead of jobs.

        Returns
        -------
        class

        Raises
        ------
        InvalidParameter
            Raised if the config file has invalid parameters.

        """
        if isinstance(filename_or_data, str):
            data = load_data(filename_or_data)
        else:
            data = filename_or_data

        data["do_not_deserialize_jobs"] = do_not_deserialize_jobs
        return cls(**data)

    def get_job(self, name):
        """Return the job matching name.

        Returns
        -------
        namedtuple

        """
        if self.get_num_jobs() == 0 and self._job_names is not None:
            # We loaded from a config file with names only.
            return self._get_job_by_name(name)

        return self._jobs.get_job(name)

    def get_num_jobs(self):
        """Return the number of jobs in the configuration.

        Returns
        -------
        int

        """
        return len(self._jobs)

    @property
    def job_global_config(self):
        """Return the global configs applied to all jobs."""
        return self._job_global_config

    def iter_jobs(self):
        """Yields a generator over all jobs.

        Yields
        ------
        iterator over JobParametersInterface

        """
        return iter(self._jobs)

    @timed_debug
    def list_jobs(self):
        """Return a list of all jobs.

        Returns
        ------
        list
            list of JobParametersInterface

        """
        return list(self.iter_jobs())

    def append_submission_group(self, submission_group):
        """Append a submission group.

        Parameters
        ----------
        submission_group : SubmissionGroup

        """
        self._submission_groups.append(submission_group)
        logger.info("Added submission group %s", submission_group.name)

    def get_default_submission_group(self):
        """Return the default submission group.

        Returns
        -------
        SubmissionGroup

        """
        name = next(iter(self.iter_jobs())).submission_group
        return self.get_submission_group(name)

    def get_submission_group(self, name):
        """Return the submission group matching name.

        Parameters
        ----------
        name : str

        Returns
        -------
        SubmissionGroup

        """
        for group in self.submission_groups:
            if group.name == name:
                return group

        raise InvalidParameter(f"submission group {name} is not stored")

    @property
    def submission_groups(self):
        """Return the submission groups.

        Returns
        -------
        list

        """
        return self._submission_groups

    @timed_debug
    def reconfigure_jobs(self, jobs):
        """Reconfigure with a list of jobs.

        Parameters
        ----------
        list of DistributionConfiguration.parameter_type

        """
        self.clear()

        for job in jobs:
            self.add_job(job)

        logger.info("Reconfigured jobs.")

    def remove_job(self, job):
        """Remove a job from the configuration.

        Parameters
        ----------
        job : JobParametersInterface

        """
        return self._jobs.remove_job(job)

    def serialize(self, include=ConfigSerializeOptions.JOBS):
        """Create data for serialization."""
        data = {
            "jobs_directory": self._jobs_directory,
            "configuration_module": self.__class__.__module__,
            "configuration_class": self.__class__.__name__,
            "format_version": self.FORMAT_VERSION,
            "user_data": self._user_data,
            "submission_groups": [x.dict() for x in self.submission_groups],
            "setup_command": self.setup_command,
            "teardown_command": self.teardown_command,
            "node_setup_command": self.node_setup_command,
            "node_teardown_command": self.node_teardown_command,
        }
        if self._job_global_config:
            data["job_global_config"] = self._job_global_config

        if self._job_post_process_config:
            data["job_post_process_config"] = self._job_post_process_config

        if include == ConfigSerializeOptions.JOBS:
            data["jobs"] = [x.serialize() for x in self.iter_jobs()]
        elif include == ConfigSerializeOptions.JOB_NAMES:
            data["job_names"] = [x.name for x in self.iter_jobs()]

        # Fill in instance-specific information.
        self._serialize(data)
        return data

    def serialize_jobs(self, directory):
        """Serializes main job data to job-specific files.

        Parameters
        ----------
        directory : str

        """
        for job in self.iter_jobs():
            basename = job.name + ".json"
            job_filename = os.path.join(directory, basename)
            dump_data(job.serialize(), job_filename, cls=ExtendedJSONEncoder)

        # We will need this to deserialize from a filename that includes only
        # job names.
        self._jobs_directory = directory

    def serialize_for_execution(self, scratch_dir, are_inputs_local=True):
        """Serialize config data for efficient execution.

        Parameters
        ----------
        scratch_dir : str
            Temporary storage space on the local system.
        are_inputs_local : bool
            Whether the existing input data is local to this system. For many
            configurations accessing the input data across the network by many
            concurrent workers can cause a bottleneck and so implementations
            may wish to copy the data locally before execution starts. If the
            storage access time is very fast the question is irrelevant.

        Returns
        -------
        str
            Name of serialized config file in scratch directory.

        """
        self._transform_for_local_execution(scratch_dir, are_inputs_local)

        # Split up the jobs to individual files so that each worker can just
        # read its own info.
        self.serialize_jobs(scratch_dir)
        data = self.serialize(ConfigSerializeOptions.JOB_NAMES)
        config_file = os.path.join(scratch_dir, CONFIG_FILE)
        dump_data(data, config_file, cls=ExtendedJSONEncoder)
        logger.info("Dumped config file locally to %s", config_file)

        return config_file

    @property
    def setup_command(self):
        """Command to run by submitter before submitting jobs"""
        return self._setup_command

    @setup_command.setter
    def setup_command(self, cmd):
        """Set command to run by submitter before submitting jobs"""
        self._setup_command = cmd

    @property
    def teardown_command(self):
        """Command to run by last node before completing jobs"""
        return self._teardown_command

    @teardown_command.setter
    def teardown_command(self, cmd):
        """Set command to run by last node before completing jobs"""
        self._teardown_command = cmd

    @property
    def node_setup_command(self):
        """Command to run on each node before starting jobs"""
        return self._node_setup_command

    @node_setup_command.setter
    def node_setup_command(self, cmd):
        """Set command to run on each node before starting jobs"""
        self._node_setup_command = cmd

    @property
    def node_teardown_command(self):
        """Command to run on each node after completing jobs"""
        return self._node_teardown_command

    @node_teardown_command.setter
    def node_teardown_command(self, cmd):
        """Set command to run on each node after completing jobs"""
        self._node_teardown_command = cmd

    def _transform_for_local_execution(self, scratch_dir, are_inputs_local):
        """Transform data for efficient execution in a local environment.
        Default implementation is a no-op. Derived classes can overridde.

        """

    def shuffle_jobs(self):
        """Shuffle the job order."""
        self._jobs.shuffle()

    def show_jobs(self):
        """Show the configured jobs."""
        for job in self.iter_jobs():
            print(job)

    def job_execution_class(self, extension_name):
        """Return the class used for job execution.

        Parameters
        ----------
        extension_name : str

        Returns
        -------
        class

        """
        return self._registry.get_extension_class(extension_name, ExtensionClassType.EXECUTION)

    def job_parameters_class(self, extension_name):
        """Return the class used for job parameters.

        Parameters
        ----------
        extension_name : str

        Returns
        -------
        class

        """
        return self._registry.get_extension_class(extension_name, ExtensionClassType.PARAMETERS)



"""
CLI to ingest results into a SQLite database.
"""

import copy
import sys
from collections import defaultdict
from datetime import datetime
from pathlib import Path

import click

from jade.common import CONFIG_FILE, OUTPUT_DIR, RESULTS_FILE, RESULTS_DB_FILE, STATS_SUMMARY_FILE
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.loggers import setup_logging
from jade.utils.sql import make_table, insert_rows
from jade.utils.utils import load_data


@click.group()
def db():
    """Database group commands"""
    setup_logging("db", None)


@click.argument("output", callback=lambda _, __, x: Path(x))
@click.command()
def ingest_results(output):
    """Ingest results from a run into a SQLite database."""
    results_file = output / RESULTS_FILE
    if not results_file.exists():
        print(f"There is no JADE results file in {output}.", file=sys.stderr)
        sys.exit(1)

    # config = create_config_from_file(output / CONFIG_FILE)
    db_file = Path(output) / RESULTS_DB_FILE
    if db_file.exists():
        db_file.unlink()

    create_result_table(db_file, results_file)

    stats_file = output / STATS_SUMMARY_FILE
    if stats_file.exists():
        create_resource_stats_table(db_file, stats_file)
    else:
        print(
            f"Cannot add resource stats to database; {stats_file} does not exist", file=sys.stderr
        )

    events_dir = output / "events"
    if events_dir.exists():
        for event_file in events_dir.iterdir():
            if event_file.suffix == ".json":
                create_event_table(db_file, event_file)
    else:
        print(f"Cannot add events to database; {events_dir} does not exist", file=sys.stderr)


def create_result_table(db_file, results_file):
    results = load_data(results_file)
    if not results:
        print(f"There are no results in {results_file}", file=sys.stderr)
        return

    rows = []
    for result in results["results"]:
        result["completion_time"] = datetime.fromtimestamp(result["completion_time"])
        rows.append(tuple(result.values()))

    for name in results["missing_jobs"]:
        row = (name, None, "missing", None, None, None)
        if len(row) != len(rows[1]):
            print(
                f"BUG: Mismatch in row length for missing job: {name}. This function may not "
                "match the definition of the Result type.",
                file=sys.stderr,
            )
            sys.exit(1)
        rows.append(row)

    table = "result"
    make_table(db_file, table, results["results"][0], primary_key="name")
    insert_rows(db_file, table, rows)


def create_resource_stats_table(db_file, stats_file):
    stats = load_data(stats_file)
    if not stats:
        print(f"There are no results in {stats_file}", file=sys.stderr)
        return

    rows_by_table = defaultdict(list)
    schema_by_table = {}
    for item in stats:
        table = item["type"].lower()
        base_row = {}
        if table == "process":
            # Make this column be listed first.
            base_row["name"] = item["name"]
        base_row["source"] = item["batch"]
        for stat_key in ("average", "minimum", "maximum"):
            stat_data = item[stat_key]
            if stat_data:
                row = copy.deepcopy(base_row)
                row["stat"] = stat_key
                for key, val in stat_data.items():
                    row[key] = val
                if table not in schema_by_table:
                    schema_by_table[table] = {}
                    fixed_columns = {}
                    for key, val in row.items():
                        for illegal in (" ", "-", "/"):
                            key = key.replace(illegal, "_")
                        fixed_columns[key] = val
                    schema_by_table[table] = fixed_columns
                rows_by_table[table].append(tuple(row.values()))

    for table, rows in rows_by_table.items():
        make_table(db_file, table, schema_by_table[table])
        insert_rows(db_file, table, rows)


def create_event_table(db_file, event_file):
    events = load_data(event_file)
    if not events:
        print(f"There are no results in {event_file}", file=sys.stderr)
        return

    rows = []
    schema = None
    table = None
    for event in events:
        event.pop("event_class")
        event.pop("category")
        data = event.pop("data")
        event.update(data)
        if schema is None:
            schema = event
            table = event["name"]
        event.pop("name")
        rows.append(tuple(event.values()))

    make_table(db_file, table, schema)
    insert_rows(db_file, table, rows)


db.add_command(ingest_results)



"""HPC management functionality"""

import logging
import os
import time

from jade.enums import Status
from jade.exceptions import InvalidParameter
from jade.hpc.common import HpcType, HpcJobStatus
from jade.hpc.fake_manager import FakeManager
from jade.hpc.local_manager import LocalManager
from jade.hpc.pbs_manager import PbsManager
from jade.hpc.slurm_manager import SlurmManager
from jade.models import HpcConfig


logger = logging.getLogger(__name__)


class HpcManager:
    """Manages HPC job submission and monitoring."""

    def __init__(self, submission_groups, output):
        self._output = output
        self._configs = {}
        self._intfs = {}
        self._hpc_type = None
        assert submission_groups
        for name, group in submission_groups.items():
            self._configs[name] = group.submitter_params.hpc_config
            self._intfs[name] = self.create_hpc_interface(group.submitter_params.hpc_config)
            if self._hpc_type is None:
                self._hpc_type = group.submitter_params.hpc_config.hpc_type

        logger.debug("Constructed HpcManager with output=%s", output)

    def _get_interface(self, submission_group_name=None):
        if submission_group_name is None:
            # In many cases we don't care which interface is used.
            # We could store job IDs by group if we need to perform actions by group
            # in the future.
            # As of now we don't track IDs at all in this class.
            return next(iter(self._intfs.values()))
        return self._intfs[submission_group_name]

    def am_i_manager(self):
        """Return True if the current node is the manager node.

        Returns
        -------
        bool

        """
        intf = self._get_interface()
        return intf.am_i_manager()

    def cancel_job(self, job_id):
        """Cancel job.

        Parameters
        ----------
        job_id : str

        Returns
        -------
        int
            return code

        """
        intf = self._get_interface()
        ret = intf.cancel_job(job_id)
        if ret == 0:
            logger.info("Successfully cancelled job ID %s", job_id)
        else:
            logger.info("Failed to cancel job ID %s", job_id)

        return ret

    def check_status(self, name=None, job_id=None):
        """Return the status of a job by name or ID.

        Parameters
        ----------
        name : str
            job name
        job_id : str
            job ID

        Returns
        -------
        HpcJobStatus

        """
        if (name is None and job_id is None) or (name is not None and job_id is not None):
            raise InvalidParameter("exactly one of name / job_id must be set")

        intf = self._get_interface()
        info = intf.check_status(name=name, job_id=job_id)
        logger.debug("info=%s", info)
        return info.status

    def check_statuses(self):
        """Check the statuses of all user jobs.

        Returns
        -------
        dict
            key is job_id, value is HpcJobStatus

        """
        intf = self._get_interface()
        return intf.check_statuses()

    def get_hpc_config(self, submission_group_name):
        """Returns the HPC interface instance.

        Parameters
        ----------
        submission_group_name : str

        Returns
        -------
        HpcManagerInterface

        """
        return self._get_interface(submission_group_name=submission_group_name)

    def get_job_stats(self, job_id):
        return self._get_interface().get_job_stats(job_id)

    def get_local_scratch(self):
        """Get path to local storage space.

        Returns
        -------
        str

        """
        return self._get_interface().get_local_scratch()

    def get_manager_node(self, job_id):
        """Return the first node in the job.

        Parameters
        ----------
        job_id : str

        Returns
        -------
        list
            list of node hostnames

        """
        intf = self._get_interface()
        return intf.list_active_nodes(job_id)[0]

    @property
    def hpc_type(self):
        """Return the type of HPC management system.

        Returns
        -------
        HpcType

        """
        return self._hpc_type

    def list_active_nodes(self, job_id):
        """Return the nodes currently participating in the job. Order should be deterministic.

        Parameters
        ----------
        job_id : str

        Returns
        -------
        list
            list of node hostnames

        """
        intf = self._get_interface()
        return intf.list_active_nodes(job_id)

    def submit(
        self,
        directory,
        name,
        script,
        submission_group_name,
        wait=False,
        keep_submission_script=True,
        dry_run=False,
    ):
        """Submits scripts to the queue for execution.

        Parameters
        ----------
        directory : str
            directory to contain the submission script
        name : str
            job name
        script : str
            Script to execute.
        submission_group_name : str
        wait : bool
            Wait for execution to complete.
        keep_submission_script : bool
            Do not delete the submission script.
        dry_run : bool
            Do not actually submit jobs. Just create the files.

        Returns
        -------
        tuple
            (job_id, submission status)

        """
        intf = self._get_interface(submission_group_name)
        intf.check_storage_configuration()

        # TODO: enable this logic if batches have unique names.
        # info = intf.check_status(name=name)
        # if info.status in (HpcJobStatus.QUEUED, HpcJobStatus.RUNNING):
        #    raise JobAlreadyInProgress(
        #        "Not submitting job '{}' because it is already active: "
        #        f"{info}"
        #    )

        filename = os.path.join(directory, name + ".sh")
        intf.create_submission_script(name, script, filename, self._output)
        logger.debug("Created submission script %s", filename)

        if dry_run:
            logger.info("Dry run mode enabled. Return without submitting.")
            return 0, Status.GOOD

        result, job_id, err = intf.submit(filename)

        if result == Status.GOOD:
            logger.info("job '%s' with ID=%s submitted successfully", name, job_id)
            if not keep_submission_script:
                os.remove(filename)
        else:
            logger.error("Failed to submit job '%s': result=%s: %s", name, result, err)

        if wait:
            self._wait_for_completion(job_id)

        return job_id, result

    @staticmethod
    def create_hpc_interface(config):
        """Returns an HPC implementation instance appropriate for the current
        environment.

        """
        if config.hpc_type == HpcType.SLURM:
            intf = SlurmManager(config)
        elif config.hpc_type == HpcType.FAKE:
            intf = FakeManager(config)
        elif config.hpc_type == HpcType.LOCAL:
            intf = LocalManager(config)
        else:
            raise ValueError("Unsupported HPC type: {}".format(config.hpc_type))

        logger.debug("HPC manager type=%s", config.hpc_type)
        return intf

    def _wait_for_completion(self, job_id):
        status = HpcJobStatus.UNKNOWN
        intf = self._get_interface()

        while status not in (HpcJobStatus.COMPLETE, HpcJobStatus.NONE):
            time.sleep(5)
            job_info = intf.check_status(job_id=job_id)
            logger.debug("job_info=%s", job_info)
            if job_info.status != status:
                logger.info("Status of job ID %s changed to %s", job_id, job_info.status)
                status = job_info.status

        logger.info("Job ID %s is complete", job_id)



from .generic_command_parameters import GenericCommandParameters
from .generic_command_configuration import GenericCommandConfiguration
from .generic_command_execution import GenericCommandExecution
from .generic_command_inputs import GenericCommandInputs



"""Provides ability to run jobs locally or on HPC."""

import datetime
import fileinput
import importlib
import logging
import os
import time
from collections import OrderedDict
from pathlib import Path

import jade
from jade.common import (
    CONFIG_FILE,
    OUTPUT_DIR,
    RESULTS_FILE,
    STATS_SUMMARY_FILE,
)
from jade.enums import Status, ResourceMonitorType
from jade.events import (
    EVENTS_FILENAME,
    EVENT_NAME_ERROR_LOG,
    StructuredLogEvent,
    EVENT_CATEGORY_ERROR,
    EVENT_CATEGORY_HPC,
    EVENT_CATEGORY_RESOURCE_UTIL,
    EVENT_NAME_BYTES_CONSUMED,
    # EVENT_NAME_SUBMIT_STARTED,
    EVENT_NAME_SUBMIT_COMPLETED,
)
from jade.exceptions import InvalidParameter
from jade.extensions.registry import Registry, ExtensionClassType
from jade.hpc.common import HpcType
from jade.hpc.hpc_manager import HpcManager
from jade.hpc.hpc_submitter import HpcSubmitter
from jade.jobs.cluster import Cluster
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.jobs.job_manager_base import JobManagerBase
from jade.jobs.job_runner import JobRunner
from jade.jobs.results_aggregator import ResultsAggregator
from jade.models import SubmitterParams
from jade.models.submission_group import make_submission_group_lookup
from jade.loggers import log_event
from jade.result import serialize_results, ResultsSummary
from jade.utils.repository_info import RepositoryInfo
from jade.utils.subprocess_manager import check_run_command, run_command
from jade.utils.utils import dump_data, get_directory_size_bytes
import jade.version


logger = logging.getLogger(__name__)


class JobSubmitter(JobManagerBase):
    """Submits jobs for execution locally or on an HPC."""

    def __init__(self, config_file, output, is_new):
        """Internal constructor. Callers should use create() or load()."""
        super().__init__(config_file, output)
        self._hpc = None
        self._is_new = is_new

    @classmethod
    def create(cls, config, output=OUTPUT_DIR):
        """Creates a new instance.

        Parameters
        ----------
        config : JobConfiguration
        output : str
            Output directory

        """
        mgr = cls(config, output, True)
        mgr.run_checks()
        config.dump(Path(output) / CONFIG_FILE, indent=2)
        return mgr

    @classmethod
    def load(cls, output):
        """Loads an instance from an existing directory."""
        config = create_config_from_file(Path(output) / CONFIG_FILE)
        return cls(config, output, False)

    def __repr__(self):
        return f"""num_jobs={self.get_num_jobs()}
results_summary={self.get_results_summmary_report()}"""

    def cancel_jobs(self, cluster):
        """Cancel running and pending jobs."""
        groups = make_submission_group_lookup(cluster.config.submission_groups)
        hpc = HpcManager(groups, self._output)
        for job_id in cluster.job_status.hpc_job_ids:
            hpc.cancel_job(job_id)

        cluster.mark_canceled()

    def submit_jobs(self, cluster, force_local=False):
        """Submit simulations. Auto-detect whether the current system is an HPC
        and submit to its queue. Otherwise, run locally.

        Parameters
        ----------
        cluster : Cluster
        force_local : bool
            If on HPC, run jobs through subprocess as if local.

        Returns
        -------
        Status

        """
        if self._is_new:
            logger.info("Submit %s jobs for execution.", self._config.get_num_jobs())
            logger.info("JADE version %s", jade.version.__version__)
            registry = Registry()
            loggers = registry.list_loggers()
            logger.debug("Registered modules for logging: %s", ", ".join(loggers))
            self._save_repository_info(registry)

            ResultsAggregator.create(self._output)

            # If an events summary file exists, it is invalid.
            events_file = os.path.join(self._output, EVENTS_FILENAME)
            if os.path.exists(events_file):
                os.remove(events_file)

            event = StructuredLogEvent(
                source="submitter",
                category=EVENT_CATEGORY_HPC,
                name=EVENT_NAME_SUBMIT_COMPLETED,
                message="job submission started",
                num_jobs=self.get_num_jobs(),
            )
            log_event(event)

            if self._config.setup_command is not None:
                env = os.environ.copy()
                env["JADE_RUNTIME_OUTPUT"] = str(self._output)
                cmd = self._config.setup_command
                logger.info("Running setup command: %s", cmd)
                check_run_command(self._config.setup_command, env=env)
        else:
            self._handle_submission_groups()

        result = Status.IN_PROGRESS
        group = self._config.get_default_submission_group()
        groups = make_submission_group_lookup(cluster.config.submission_groups)
        self._hpc = HpcManager(groups, self._output)

        if self._hpc.hpc_type == HpcType.LOCAL or force_local:
            runner = JobRunner(self._config, output=self._output)
            num_parallel_processes_per_node = (
                group.submitter_params.num_parallel_processes_per_node
            )
            verbose = group.submitter_params.verbose
            result = runner.run_jobs(
                verbose=verbose, num_parallel_processes_per_node=num_parallel_processes_per_node
            )
            agg = ResultsAggregator.load(self._output)
            agg.process_results()
            is_complete = True
        else:
            is_complete = self._submit_to_hpc(cluster)

        if is_complete:
            result = self._handle_completion(cluster)

        return result

    def _handle_completion(self, cluster):
        result = Status.GOOD
        self._results = ResultsAggregator.list_results(self._output)
        if len(self._results) != self._config.get_num_jobs():
            finished_jobs = {x.name for x in self._results}
            all_jobs = {x.name for x in self._config.iter_jobs()}
            missing_jobs = sorted(all_jobs.difference(finished_jobs))
            logger.error(
                "Error in result totals. num_results=%s total_num_jobs=%s",
                len(self._results),
                self._config.get_num_jobs(),
            )
            logger.error(
                "These jobs did not finish: %s. Check for process crashes or HPC timeouts.",
                missing_jobs,
            )
            result = Status.ERROR
        else:
            missing_jobs = []

        self.write_results_summary(RESULTS_FILE, missing_jobs)

        if self._config.teardown_command is not None:
            env = os.environ.copy()
            env["JADE_RUNTIME_OUTPUT"] = str(self._output)
            cmd = self._config.teardown_command
            logger.info("Running teardown command: %s", cmd)
            start = time.time()
            ret = run_command(self._config.teardown_command, env=env)
            if ret != 0:
                logger.error("Failed to run teardown command: %s", ret)
            logger.info("Teardown command duration = %s seconds", time.time() - start)

        self._log_error_log_messages(self._output)

        bytes_consumed = get_directory_size_bytes(self._output, recursive=False)
        event = StructuredLogEvent(
            source="submitter",
            category=EVENT_CATEGORY_RESOURCE_UTIL,
            name=EVENT_NAME_BYTES_CONSUMED,
            message="main output directory size",
            bytes_consumed=bytes_consumed,
        )
        log_event(event)

        event = StructuredLogEvent(
            source="submitter",
            category=EVENT_CATEGORY_HPC,
            name=EVENT_NAME_SUBMIT_COMPLETED,
            message="job submission completed",
            num_jobs=self.get_num_jobs(),
        )
        log_event(event)

        group = self._config.get_default_submission_group()
        if group.submitter_params.generate_reports:
            self.generate_reports(self._output, group.submitter_params.resource_monitor_type)
        cluster.mark_complete()

        if cluster.config.pipeline_stage_num is not None:
            # The pipeline directory must be the one above this one.
            pipeline_dir = os.path.dirname(self._output)
            next_stage = cluster.config.pipeline_stage_num + 1
            cmd = (
                f"jade pipeline submit-next-stage {pipeline_dir} "
                f"--stage-num={next_stage} "
                f"--return-code={result.value}"
            )
            run_command(cmd)

        return result

    def write_results_summary(self, filename, missing_jobs):
        """Write the results to filename in the output directory."""
        data = OrderedDict()
        data["jade_version"] = jade.version.__version__
        now = datetime.datetime.now()
        data["timestamp"] = now.strftime("%m/%d/%Y %H:%M:%S")
        data["base_directory"] = os.getcwd()
        results = self._build_results(missing_jobs)
        data["results_summary"] = results["summary"]
        data["missing_jobs"] = missing_jobs
        data["results"] = results["results"]

        output_file = os.path.join(self._output, filename)
        dump_data(data, output_file)

        logger.info("Wrote results to %s.", output_file)
        num_successful = results["summary"]["num_successful"]
        num_canceled = results["summary"]["num_canceled"]
        num_failed = results["summary"]["num_failed"]
        num_missing = len(missing_jobs)
        total = num_successful + num_failed + num_missing
        log_func = logger.info if num_successful == total else logger.warning
        log_func(
            "Successful=%s Failed=%s Canceled=%s Missing=%s Total=%s",
            num_successful,
            num_failed,
            num_canceled,
            num_missing,
            total,
        )

        return output_file

    def _build_results(self, missing_jobs):
        num_successful = 0
        num_failed = 0
        num_canceled = 0
        for result in self._results:
            if result.is_successful():
                num_successful += 1
            elif result.is_failed():
                num_failed += 1
            else:
                assert result.is_canceled(), str(result)
                num_canceled += 1

        return {
            "results": serialize_results(self._results),
            "summary": {
                "num_successful": num_successful,
                "num_failed": num_failed,
                "num_canceled": num_canceled,
                "num_missing": len(missing_jobs),
            },
        }

    def _save_repository_info(self, registry):
        extensions = registry.list_extensions()
        extension_packages = set(["jade"])
        for ext in extensions:
            exec_module = ext[ExtensionClassType.EXECUTION].__module__
            name = exec_module.split(".")[0]
            extension_packages.add(name)

        for name in extension_packages:
            try:
                package = importlib.import_module(name)
                repo_info = RepositoryInfo(package)
                patch = os.path.join(self._output, f"{name}-diff.patch")
                repo_info.write_diff_patch(patch)
                logger.info("%s repository information: %s", name, repo_info.summary())
            except InvalidParameter:
                pass

    @staticmethod
    def _log_error_log_messages(directory):
        for event in JobSubmitter.find_error_log_messages(directory):
            log_event(event)

    @staticmethod
    def find_error_log_messages(directory):
        """Parse output log files for error messages

        Parameters
        ----------
        directory : str
            output directory

        """
        substrings = (
            "DUE TO TIME LIMIT",  # includes slurmstepd, but check this first
            "srun",
            "slurmstepd",
            "Traceback",
        )

        filenames = [os.path.join(directory, x) for x in os.listdir(directory) if x.endswith(".e")]

        if not filenames:
            return

        for line in fileinput.input(filenames):
            for substring in substrings:
                if substring in line:
                    event = StructuredLogEvent(
                        source="submitter",
                        category=EVENT_CATEGORY_ERROR,
                        name=EVENT_NAME_ERROR_LOG,
                        message="Detected error message in log.",
                        error=substring,
                        filename=fileinput.filename(),
                        line_number=fileinput.lineno(),
                        text=line.strip(),
                    )
                    yield event
                    # Only find one match in a single line.
                    break

    @staticmethod
    def generate_reports(directory, resource_monitor_type):
        """Create reports summarizing the output results of a set of jobs.

        Parameters
        ----------
        directory : str
            output directory
        resource_monitor_type : ResourceMonitorType

        """
        commands = [
            (f"jade show-results -o {directory}", "results.txt"),
            (f"jade show-events -o {directory} --categories Error", "errors.txt"),
        ]
        if resource_monitor_type != ResourceMonitorType.NONE:
            commands.append((f"jade stats show -o {directory} --summary-only", "stats.txt"))
            commands.append((f"jade stats show -o {directory} -j", STATS_SUMMARY_FILE))
        if resource_monitor_type == ResourceMonitorType.PERIODIC:
            commands.append((f"jade stats plot -o {directory}", None))
        # This must come after stats commands.
        commands.append((f"jade db ingest-results {directory}", None))

        reports = []
        for cmd in commands:
            output = {}
            ret = run_command(cmd[0], output=output)
            if ret != 0:
                logger.error("Failed to run [%s]: %s: %s", cmd[0], ret, output["stderr"])
                continue

            if cmd[1] is not None:
                filename = os.path.join(directory, cmd[1])
                with open(filename, "w") as f_out:
                    if "json" not in cmd[1]:
                        f_out.write(cmd[0] + "\n\n")
                    f_out.write(output["stdout"])
                    reports.append(filename)
                print(f"wrote output to {filename}")

        logger.info("Generated reports %s.", " ".join(reports))

        return 0

    def _submit_to_hpc(self, cluster):
        hpc_submitter = HpcSubmitter(
            self._config,
            self._config_file,
            cluster,
            self._output,
        )

        if hpc_submitter.run():
            logger.info("All submitters have completed.")
            return True

        logger.debug("jobs are still pending")
        return False

    def run_checks(self):
        """Checks the configuration for errors. May mutate the config."""
        self._config.check_submission_groups()
        for group in self._config.submission_groups:
            if group.submitter_params.per_node_batch_size == 0:
                self._config.check_job_estimated_run_minutes(group.name)
        self._config.check_job_dependencies()
        self._config.check_job_runtimes()
        self._config.check_spark_config()

    @staticmethod
    def run_submit_jobs(config, output, local=False, dry_run=False, pipeline_stage_num=None):
        """Allows submission from an existing Python process."""
        os.makedirs(output, exist_ok=True)

        mgr = JobSubmitter.create(config, output=output)
        cluster = Cluster.create(
            output,
            mgr.config,
            pipeline_stage_num=pipeline_stage_num,
        )

        ret = 1
        try:
            status = mgr.submit_jobs(cluster, force_local=local)
            if status == Status.IN_PROGRESS:
                check_cmd = f"jade show-status -o {output}"
                if not dry_run:
                    print(f"Jobs are in progress. Run '{check_cmd}' for updates.")
                ret = 0
            else:
                ret = status.value
        finally:
            cluster.demote_from_submitter()
            if local:
                # These files were not used in this case.
                cluster.delete_files_internal()

        return ret



"""Common definitions for the jade package."""


ANALYSIS_DIR = "analysis"
CONFIG_FILE = "config.json"
DEFAULT_SUBMISSION_GROUP = "default"
EVENTS_DIR = "events"
HPC_CONFIG_FILE = "hpc_config.toml"
JOBS_OUTPUT_DIR = "job-outputs"
JOBS_STDIO_DIR = "job-stdio"
OUTPUT_DIR = "output"
POST_PROCESSING_CONFIG_FILE = "post-config.json"
RESULTS_DB_FILE = "results.db"
RESULTS_DIR = "results"
RESULTS_FILE = "results.json"
SCRIPTS_DIR = "scripts"
STATS_DIR = "stats"
STATS_SUMMARY_FILE = "stats_summary.json"



"""Manages extensions registered with JADE."""

import copy
import enum
import importlib
import logging
import os
import pathlib

from jade.exceptions import InvalidParameter
from jade.utils.utils import dump_data, load_data


DEFAULT_REGISTRY = {
    "extensions": [
        {
            "name": "generic_command",
            "description": "Allows batching of a list of CLI commands.",
            "job_execution_module": "jade.extensions.generic_command.generic_command_execution",
            "job_execution_class": "GenericCommandExecution",
            "job_configuration_module": "jade.extensions.generic_command.generic_command_configuration",
            "job_configuration_class": "GenericCommandConfiguration",
            "job_parameters_module": "jade.extensions.generic_command.generic_command_parameters",
            "job_parameters_class": "GenericCommandParameters",
            "cli_module": "jade.extensions.generic_command.cli",
        },
    ],
    "logging": [
        "jade",
    ],
}


class ExtensionClassType(enum.Enum):
    """Possible values for computational sequencing mode"""

    CLI = "cli_module"
    CONFIGURATION = "config_class"
    EXECUTION = "exec_class"
    PARAMETERS = "param_class"


logger = logging.getLogger(__name__)


class Registry:
    """Manages extensions registered with JADE."""

    _REGISTRY_FILENAME = ".jade-registry.json"
    FORMAT_VERSION = "v0.2.0"

    def __init__(self, registry_filename=None):
        if registry_filename is None:
            if "JADE_REGISTRY" in os.environ:
                self._registry_filename = os.environ["JADE_REGISTRY"]
            else:
                self._registry_filename = os.path.join(
                    str(pathlib.Path.home()),
                    self._REGISTRY_FILENAME,
                )
        else:
            self._registry_filename = registry_filename

        self._extensions = {}
        self._loggers = set()
        if not os.path.exists(self._registry_filename):
            self.reset_defaults()
        else:
            data = self._check_registry_config(self._registry_filename)
            for extension in data["extensions"]:
                self._add_extension(extension)
            for package_name in data["logging"]:
                self._loggers.add(package_name)

    def _add_extension(self, extension):
        for field in DEFAULT_REGISTRY["extensions"][0]:
            if field not in extension:
                raise InvalidParameter(f"required field {field} not present")

        try:
            cmod = importlib.import_module(extension["job_configuration_module"])
            emod = importlib.import_module(extension["job_execution_module"])
            pmod = importlib.import_module(extension["job_parameters_module"])
            cli_mod = importlib.import_module(extension["cli_module"])
        except ImportError as exc:
            if "statsmodels" in exc.msg:
                # Older versions of Jade installed the demo extension into the registry as
                # well as its dependencies. Newer versions do not. This causes import errors
                # when a user upgrades to the newer version.
                # Remove the demo extension. The user can add it later if they want.
                # This can be removed whenever all users have gone through an upgrade.
                self._remove_demo_extension()
                return
            else:
                raise

        ext = copy.copy(extension)
        ext[ExtensionClassType.CONFIGURATION] = getattr(cmod, extension["job_configuration_class"])
        ext[ExtensionClassType.EXECUTION] = getattr(emod, extension["job_execution_class"])
        ext[ExtensionClassType.PARAMETERS] = getattr(pmod, extension["job_parameters_class"])
        ext[ExtensionClassType.CLI] = cli_mod

        self._extensions[extension["name"]] = ext

    def _check_registry_config(self, filename):
        data = load_data(filename)
        if isinstance(data, list):
            # Workaround to support the old registry format. 03/06/2020
            # It can be removed eventually.
            new_data = {
                "extensions": data,
                "logging": DEFAULT_REGISTRY["logging"],
            }
            dump_data(new_data, self.registry_filename, indent=4)
            print(
                "\nReformatted registry. Refer to `jade extensions --help` "
                "for instructions on adding logging for external packages.\n"
            )
            data = new_data

        format = data.get("format_version", "v0.1.0")
        if format == "v0.1.0":
            self.reset_defaults()
            data = load_data(filename)
            print(
                "\nWARNING: Reformatted registry. You will need to "
                "re-register any external extensions.\n"
            )
        return data

    def _serialize_registry(self):
        data = {
            "extensions": [],
            "logging": list(self._loggers),
            "format_version": self.FORMAT_VERSION,
        }
        for _, extension in sorted(self._extensions.items()):
            ext = {k: v for k, v in extension.items() if not isinstance(k, ExtensionClassType)}
            data["extensions"].append(ext)

        filename = self.registry_filename
        dump_data(data, filename, indent=4)
        logger.debug("Serialized data to %s", filename)

    def add_logger(self, package_name):
        """Add a package name to the logging registry.

        Parameters
        ----------
        package_name : str

        """
        self._loggers.add(package_name)
        self._serialize_registry()

    def remove_logger(self, package_name):
        """Remove a package name from the logging registry.

        Parameters
        ----------
        package_name : str

        """
        self._loggers.remove(package_name)
        self._serialize_registry()

    def list_loggers(self):
        """List the package names registered to be logged.

        Returns
        -------
        list

        """
        return sorted(list(self._loggers))

    def show_loggers(self):
        """Print the package names registered to be logged."""
        print(", ".join(self.list_loggers()))

    def get_extension_class(self, extension_name, class_type):
        """Get the class associated with the extension.

        Parameters
        ----------
        extension_name : str
        class_type : ExtensionClassType

        Raises
        ------
        InvalidParameter
            Raised if the extension is not registered.

        """
        extension = self._extensions.get(extension_name)
        if extension is None:
            raise InvalidParameter(f"{extension_name} is not registered")

        return extension[class_type]

    def is_registered(self, extension_name):
        """Check if the extension is registered"""
        return extension_name in self._extensions

    def iter_extensions(self):
        """Return an iterator over registered extensions.

        Returns
        -------
        dict_values

        """
        return self._extensions.values()

    def list_extensions(self):
        """Return a list of registered extensions.

        Returns
        -------
        list of dict

        """
        return list(self.iter_extensions())

    def register_extension(self, extension):
        """Registers an extension in the registry.

        Parameters
        ----------
        extension : dict

        Raises
        ------
        InvalidParameter
            Raised if the extension is invalid.

        """

        self._add_extension(extension)
        self._serialize_registry()
        logger.debug("Registered extension %s", extension["name"])

    @property
    def registry_filename(self):
        """Return the filename that stores the registry."""
        return self._registry_filename

    def reset_defaults(self):
        """Reset the registry to its default values."""
        self._extensions.clear()
        self._loggers.clear()
        for extension in DEFAULT_REGISTRY["extensions"]:
            self.register_extension(extension)
        for package_name in DEFAULT_REGISTRY["logging"]:
            self.add_logger(package_name)
        self._serialize_registry()

        logger.debug("Initialized registry to its defaults.")

    def show_extensions(self):
        """Show the registered extensions."""
        print("JADE Extensions:")
        for name, extension in sorted(self._extensions.items()):
            print(f"  {name}:  {extension['description']}")

    def unregister_extension(self, extension_name):
        """Unregisters an extension.

        Parameters
        ----------
        extension_name : str

        """
        if extension_name not in self._extensions:
            raise InvalidParameter(f"extension {extension_name} isn't registered")

        self._extensions.pop(extension_name)
        self._serialize_registry()

    def register_demo_extension(self):
        self.register_extension(
            {
                "name": "demo",
                "description": "Country based GDP auto-regression analysis",
                "job_execution_module": "jade.extensions.demo.autoregression_execution",
                "job_execution_class": "AutoRegressionExecution",
                "job_configuration_module": "jade.extensions.demo.autoregression_configuration",
                "job_configuration_class": "AutoRegressionConfiguration",
                "job_parameters_module": "jade.extensions.demo.autoregression_parameters",
                "job_parameters_class": "AutoRegressionParameters",
                "cli_module": "jade.extensions.demo.cli",
            },
        )

    def _remove_demo_extension(self):
        registry_file = pathlib.Path.home() / self._REGISTRY_FILENAME
        if not registry_file.exists():
            return

        data = load_data(registry_file)
        found = False
        for i, ext in enumerate(data["extensions"]):
            if ext["name"] == "demo":
                data["extensions"].pop(i)
                found = True
                break
        if found:
            dump_data(data, registry_file, indent=2)



"""Defines the cluster configuration"""

from typing import List, Optional

from pydantic import Field

from jade.models import JadeBaseModel, SubmissionGroup


class ClusterConfig(JadeBaseModel):
    """Describes the roles of compute nodes participating in a JADE config."""

    submitter: Optional[str] = Field(
        title="submitter",
        description="defines the current submitter, hostname or None",
        default=None,
    )
    submission_groups: List[SubmissionGroup] = Field(
        title="submission_groups",
        description="defines the submitter options selected by the user",
    )
    path: str = Field(
        title="path",
        description="directory on shared filesystem containing config",
    )
    pipeline_stage_num: Optional[int] = Field(
        title="pipeline_stage_num",
        description="stage number if the config is part of a pipeline",
    )
    num_jobs: int = Field(
        title="num_jobs",
        description="total number of jobs in configuration",
    )
    submitted_jobs: int = Field(
        title="submitted_jobs",
        description="number of jobs submitted (never decrements)",
        default=0,
    )
    completed_jobs: int = Field(
        title="completed_jobs",
        description="number of jobs that have completed",
        default=0,
    )
    is_complete: bool = Field(
        title="is_complete",
        description="set to True when the submission is complete",
        default=False,
    )
    is_canceled: bool = Field(
        title="is_canceled",
        description="set to True if the submission is canceled",
        default=False,
    )
    version: int = Field(
        title="version",
        description="version of the config, increments with each update",
    )



from pathlib import Path
from typing import Dict, List, Set

from pydantic import Field, root_validator

from jade.models import JadeBaseModel


class SparkContainerModel(JadeBaseModel):
    """Model definition for a Spark container"""

    path: str = Field(
        title="path",
        description="Path to container that can run Spark",
    )


class SparkConfigModel(JadeBaseModel):
    """Model definition for a Spark configuration"""

    collect_worker_logs: bool = Field(
        title="collect_worker_logs",
        description="Collect logs from worker processes.",
        default=False,
    )
    conf_dir: str = Field(
        title="conf_dir",
        description="Spark configuration directory",
        default="spark-conf",
    )
    container: SparkContainerModel = Field(
        title="container",
        description="Container parameters",
    )
    enabled: bool = Field(
        title="enabled",
        description="Set to true to run this job on a Spark cluster",
        default=False,
    )
    master_node_memory_overhead_gb: int = Field(
        title="master_node_memory_overhead_gb",
        description="Memory overhead for Spark master processes",
        default=3,
    )
    node_memory_overhead_gb: int = Field(
        title="node_memory_overhead_gb",
        description="Memory overhead for node operating system and existing applications",
        default=10,
    )
    run_user_script_inside_container: bool = Field(
        title="run_user_script_inside_container",
        description="Run the user script inside of the container.",
        default=False,
    )
    use_tmpfs_for_scratch: bool = Field(
        title="use_tmpfs_for_scratch",
        description="Use node's tmpfs instead of internal storage for scratch space.",
        default=False,
    )
    alt_scratch: str = Field(
        title="alt_scratch",
        description="Use this alternative directory for scratch space.",
        default=None,
    )
    worker_memory_gb: int = Field(
        title="worker_memory_gb",
        description="If 0, give all node memory minus overhead to worker.",
        default=0,
    )

    @root_validator(pre=True)
    def handle_legacy_values(cls, values):
        run_outside = values.pop("run_user_script_outside_container", None)
        if run_outside is not None and "run_user_script_inside_container" not in values:
            values["run_user_script_inside_container"] = not run_outside
        return values

    def get_spark_script(self):
        wrapper = Path(self.conf_dir) / "bin" / "run_spark_script_wrapper.sh"
        script = Path(self.conf_dir) / "bin" / "run_spark_script.sh"
        return f"{wrapper} {script}"

    def get_start_master(self):
        return self.get_spark_script() + " sbin/start-master.sh"

    def get_stop_master(self):
        return self.get_spark_script() + " sbin/stop-master.sh"

    def get_start_worker(self, memory, cluster):
        return self.get_spark_script() + f" sbin/start-worker.sh -m {memory} {cluster}"

    def get_stop_worker(self):
        return self.get_spark_script() + " sbin/stop-worker.sh"

    def get_start_history_server(self):
        return self.get_spark_script() + " sbin/start-history-server.sh"

    def get_stop_history_server(self):
        return self.get_spark_script() + " sbin/stop-history-server.sh"

    def get_run_user_script(self):
        assert self.run_user_script_inside_container
        return Path(self.conf_dir) / "bin" / "run_user_script_wrapper.sh"



"""Common definitions for HPC functionality"""

from collections import namedtuple
import enum


class HpcJobStatus(enum.Enum):
    """Represents the status of an HPC job."""

    UNKNOWN = "unknown"
    NONE = "none"
    QUEUED = "queued"
    RUNNING = "running"
    COMPLETE = "complete"


HpcJobInfo = namedtuple("HpcJobInfo", "job_id, name, status")
HpcJobStats = namedtuple(
    "HpcJobStats", "hpc_job_id, name, start, end, state, account, partition, qos"
)


class HpcQos(enum.Enum):
    """HPC Quality of Service values"""

    LOW = 0
    HIGH = 1


class HpcType(enum.Enum):
    """HPC types"""

    LOCAL = "local"
    PBS = "pbs"
    SLURM = "slurm"
    FAKE = "fake"



"""Enums for the jade package."""

import enum
import re

import toml


class Status(enum.Enum):
    """Return status."""

    GOOD = 0
    ERROR = 1
    IN_PROGRESS = 2


class Mode(enum.Enum):
    """Possible values for computational sequencing mode"""

    PARALLEL = "parallel"
    SEQUENTIAL = "sequential"


class JobCompletionStatus(enum.Enum):
    """Possible values for job completion status"""

    FINISHED = "finished"  # Job ran. return code could be 0 or 1
    CANCELED = "canceled"  # Job never ran.
    MISSING = "missing"  # No result recorded. Happens with walltime timeout.


class ResourceMonitorType(enum.Enum):
    """Type of resource monitoring to perform"""

    AGGREGATION = "aggregation"
    PERIODIC = "periodic"
    NONE = "none"


PUBLIC_ENUMS = {
    "Mode": Mode,
    "JobCompletionStatus": JobCompletionStatus,
}


_REGEX_ENUM = re.compile(r"^(\w+)\.(\w+)$")


def get_enum_from_str(string):
    """Converts a enum that's been written as a string back to an enum.

    Parameters
    ----------
    string : str
        string to convert

    Returns
    -------
    tuple
        converted, converted enum or original

    """
    converted = False
    match = _REGEX_ENUM.search(string)
    if match and match.group(1) in PUBLIC_ENUMS:
        obj = getattr(PUBLIC_ENUMS[match.group(1)], match.group(2))
        converted = True
    else:
        obj = string

    return converted, obj


def get_enum_from_value(cls, value):
    """Gets the enum for the given value."""
    for enum_ in cls:
        if enum_.value == value:
            return enum_
    raise Exception("Unknown value: {} {}".format(cls, value))


class EnumEncoder(toml.TomlEncoder):
    """Custom encoder for enums."""

    def __init__(self, _dict=dict, preserve=False):
        super(EnumEncoder, self).__init__(_dict, preserve)
        self.dump_funcs[enum.Enum] = EnumEncoder.dump_enum

    @staticmethod
    def dump_enum(val):
        """Return the enum value converted to string."""
        return str(val.value)



"""Contains common functionality for managing jobs."""

import abc
import json
import logging
import os
from pathlib import Path

from jade.jobs.cluster import Cluster
from jade.common import JOBS_OUTPUT_DIR, JOBS_STDIO_DIR, RESULTS_DIR, STATS_DIR, CONFIG_FILE
from jade.jobs.job_configuration_factory import create_config_from_file


logger = logging.getLogger(__name__)


class JobManagerBase(abc.ABC):
    """Base class for managing jobs."""

    def __init__(self, config, output_dir):
        self._config = config
        self._config_file = Path(output_dir) / CONFIG_FILE
        self._output = output_dir
        self._jobs_output = os.path.join(self._output, JOBS_OUTPUT_DIR)
        self._jobs_stdio_dir = os.path.join(self._output, JOBS_STDIO_DIR)
        self._results_dir = os.path.join(self._output, RESULTS_DIR)
        self._stats_dir = os.path.join(self._output, STATS_DIR)
        self._results = []  # contains Result objects

        os.makedirs(self._output, exist_ok=True)
        os.makedirs(self._jobs_output, exist_ok=True)
        os.makedirs(self._jobs_stdio_dir, exist_ok=True)
        os.makedirs(self._results_dir, exist_ok=True)
        os.makedirs(self._stats_dir, exist_ok=True)

    @property
    def config(self):
        return self._config

    def get_num_jobs(self):
        """Get the number of jobs to execute.

        Returns
        -------
        int

        """
        return self._config.get_num_jobs()

    def get_completed_results(self):
        """Get results of completed jobs.

        Returns
        -------
        list of Result

        """
        return self._results[:]

    def get_results_summmary(self):
        """Get a job results.

        Returns
        -------
        dict

        """
        num_successful = 0
        num_failed = 0
        for job in self._results:
            if job.status == 0:
                num_successful += 1
            else:
                num_failed += 1

        results = {
            "num_jobs": self.get_num_jobs(),
            "num_complete": len(self._results),
            "num_successful": num_successful,
            "num_failed": num_failed,
        }

        return results

    def get_results_summmary_report(self):
        """Get a summary of job results in text form.

        Returns
        -------
        str

        """
        data = self.get_results_summmary()
        return json.dumps(data, indent=4)

    def get_batch_post_process_config(self):
        """Get the batch post-process config data"""
        return self._config.batch_post_process_config

    def _handle_submission_groups(self):
        # The JobConfiguration will not have any groups if the user didn't define any.
        # Reload from the cluster config.
        if not self._config.submission_groups:
            submission_groups = Cluster.deserialize_submission_groups(Path(self._output))
            assert len(submission_groups) == 1
            self._config.append_submission_group(submission_groups[0])



"""SLURM management functionality"""

import logging
import multiprocessing
import tempfile

from jade.enums import Status
from jade.hpc.common import HpcJobStatus, HpcJobInfo
from jade.hpc.hpc_manager_interface import HpcManagerInterface


logger = logging.getLogger(__name__)


DEFAULTS = {
    "walltime": 60 * 12,
    "interface": "ib0",
    "local_directory": tempfile.gettempdir(),
    "memory": 5000,
}


class LocalManager(HpcManagerInterface):
    """Manages local execution of jobs."""

    _STATUSES = {
        "PD": HpcJobStatus.QUEUED,
        "R": HpcJobStatus.RUNNING,
        "CG": HpcJobStatus.COMPLETE,
    }

    def __init__(self, _):
        pass

    def am_i_manager(self):
        return True

    def cancel_job(self, job_id):
        return 0

    def check_status(self, name=None, job_id=None):
        return HpcJobInfo("", "", HpcJobStatus.NONE)

    def check_statuses(self):
        return {}

    def check_storage_configuration(self):
        pass

    def create_cluster(self):
        pass

    def create_local_cluster(self):
        pass

    def create_submission_script(self, name, script, filename, path):
        pass

    def get_config(self):
        return {"hpc": {}}

    def get_current_job_id(self):
        return None

    def get_job_stats(self, job_id):
        return None

    def get_local_scratch(self):
        return tempfile.gettempdir()

    def get_node_id(self):
        return "0"

    @staticmethod
    def get_num_cpus():
        return multiprocessing.cpu_count()

    def list_active_nodes(self, job_id):
        assert False

    def log_environment_variables(self):
        pass

    def submit(self, filename):
        return Status.GOOD, 1, None



"""Defines JADE base model"""

from pathlib import Path

from pydantic import BaseModel

from jade.utils.utils import load_data


class JadeBaseModel(BaseModel):
    """Base class for JADE models."""

    class Config:
        title = "JadeBaseModel"
        anystr_strip_whitespace = True
        validate_assignment = True
        validate_all = True
        extra = "forbid"
        use_enum_values = False
        allow_population_by_field_name = True

    @classmethod
    def load(cls, path: Path):
        """Load a model from a file."""
        return cls(**load_data(path))



"""Common exceptions for jade package"""


class JadeBaseException(Exception):
    """All JADE exceptions should derive from this class."""


class ExecutionError(JadeBaseException):
    """Raised when execution fails."""


class InvalidConfiguration(JadeBaseException):
    """Raised when the configuration is invalid."""


class InvalidParameter(JadeBaseException):
    """Raised when bad user input is detected."""


class JobAlreadyInProgress(JadeBaseException):
    """Raised when a local job is started while another is in progress."""


class UserAbort(JadeBaseException):
    """Raised when the user has aborted the operation."""


class InvalidExtension(JadeBaseException):
    """Raise when extension does not exist in EXTENSIONS"""



"""Models for submitter options"""

import re
from datetime import timedelta
from typing import Optional

from pydantic import Field

from jade.enums import ResourceMonitorType
from jade.models import JadeBaseModel, HpcConfig, SingularityParams


class ResourceMonitorStats(JadeBaseModel):
    """Defines the stats to monitor."""

    cpu: bool = Field(
        description="Monitor CPU utilization",
        default=True,
    )
    disk: bool = Field(
        description="Monitor disk/storage utilization",
        default=False,
    )
    memory: bool = Field(
        description="Monitor memory utilization",
        default=True,
    )
    network: bool = Field(
        description="Monitor network utilization",
        default=False,
    )
    process: bool = Field(
        description="Monitor per-job process utilization",
        default=False,
    )
    include_child_processes: bool = Field(
        description="Include stats from direct child processes in utilization for each job.",
        default=True,
    )
    recurse_child_processes: bool = Field(
        description="Recurse child processes to find all descendants.",
        default=False,
    )


class SubmitterParams(JadeBaseModel):
    """Defines the submitter options selected by the user."""

    generate_reports: bool = Field(
        description="Controls whether to generate reports after completion",
        default=True,
    )
    hpc_config: HpcConfig = Field(
        description="HPC config options",
    )
    max_nodes: Optional[int] = Field(
        description="Max number of compute nodes to use simultaneously, default is unbounded",
        default=None,
    )
    num_parallel_processes_per_node: Optional[int] = Field(
        description="Number of processes to run in parallel on each node",
        default=None,
        alias="num_processes",
    )
    per_node_batch_size: int = Field(
        description="How many jobs to assign to each node",
        default=500,
    )
    # The next two parameters are obsolete and will eventually be deleted.
    node_setup_script: Optional[str] = Field(
        description="Script to run on each node before starting jobs",
        default=None,
    )
    node_shutdown_script: Optional[str] = Field(
        description="Script to run on each node after completing jobs",
        default=None,
    )
    poll_interval: int = Field(
        description="Interval in seconds on which to poll jobs for status",
        default=10,
    )
    resource_monitor_interval: Optional[int] = Field(
        description="Interval in seconds on which to collect resource stats. Disable monitoring "
        "by setting this to None/null."
        "summaries of stats.",
        default=10,
    )
    resource_monitor_type: ResourceMonitorType = Field(
        description=f"Type of resource monitoring to perform. Options: {[x.value for x in ResourceMonitorType]}",
        default=ResourceMonitorType.AGGREGATION,
    )
    resource_monitor_stats: ResourceMonitorStats = Field(
        description="Resource utilization stats to monitor",
        default=ResourceMonitorStats(),
    )
    try_add_blocked_jobs: bool = Field(
        description="Add blocked jobs to a batch if all blocking jobs are in the batch. "
        "Be aware of time constraints.",
        default=True,
    )
    time_based_batching: bool = Field(
        description="Use time-based batching instead of job-count-based batching",
        default=False,
    )
    dry_run: bool = Field(
        description="Dry run mode; don't start any jobs",
        default=False,
    )
    verbose: bool = Field(
        description="Enable debug logging",
        default=False,
    )
    singularity_params: Optional[SingularityParams] = Field(
        description="Singularity container parameters",
        default=None,
    )
    distributed_submitter: bool = Field(
        description="Submit new jobs and update status on compute nodes.",
        default=True,
    )

    def get_wall_time(self):
        """Return the wall time from the HPC parameters.

        Returns
        -------
        timedelta

        """
        wall_time = getattr(self.hpc_config.hpc, "walltime", None)
        if wall_time is None:
            return timedelta(seconds=0xFFFFFFFF)  # largest 8-byte integer
        return _to_timedelta(wall_time)

    def dict(self, *args, **kwargs):
        data = super().dict(*args, **kwargs)
        if data["node_setup_script"] is None:
            data.pop("node_setup_script")
        if data["node_shutdown_script"] is None:
            data.pop("node_shutdown_script")
        return data


_REGEX_WALL_TIME = re.compile(r"(\d+):(\d+):(\d+)")


def _to_timedelta(wall_time):
    match = _REGEX_WALL_TIME.search(wall_time)
    assert match
    hours = int(match.group(1))
    minutes = int(match.group(2))
    seconds = int(match.group(3))
    return timedelta(hours=hours, minutes=minutes, seconds=seconds)



"""Models for HPC configurations"""

import re
from typing import Optional, Union, List

from pydantic import Field, validator, root_validator, validator

from jade.hpc.common import HpcType
from jade.models.base import JadeBaseModel


class SlurmConfig(JadeBaseModel):
    """Defines config options for the SLURM queueing system."""

    account: str = Field(
        title="account",
        description="Project account to use",
    )
    partition: Optional[str] = Field(
        title="partition",
        description="HPC partition on which to submit",
        default=None,
    )
    reservation: Optional[str] = Field(
        title="reservation",
        description="HPC reservation on which to submit",
        default=None,
    )
    qos: Optional[str] = Field(
        title="qos",
        description="Set to high to get faster node allocations at twice the cost",
        default=None,
    )
    walltime: str = Field(
        title="walltime",
        description="Maximum time allocated to each node",
        default="4:00:00",
    )
    gres: Optional[str] = Field(
        title="gpu",
        description="Request nodes that have at least this number of GPUs. Ex: 'gpu:2'",
    )
    mem: Optional[str] = Field(
        title="mem",
        description="Request nodes that have at least this amount of memory",
        default=None,
    )
    tmp: Optional[str] = Field(
        title="tmp",
        description="Request nodes that have at least this amount of storage scratch space",
        default=None,
    )
    nodes: Optional[int] = Field(
        title="nodes",
        description="Number of nodes to use for each job",
        default=None,
    )
    ntasks: Optional[int] = Field(
        title="ntasks",
        description="Number of tasks per job (nodes is not required if this is provided)",
        default=None,
    )
    ntasks_per_node: Optional[int] = Field(
        title="ntasks_per_node",
        description="Number of tasks per job (max in number of CPUs)",
        default=None,
    )

    @validator("gres")
    def check_gpus(cls, gres):
        if gres is None:
            return gres
        if re.search(r"^gpu:(\d+)$", gres) is None:
            raise ValueError(
                "gres value must follow the format 'gres=gpu:N' where N is the number of required GPUs"
            )
        return gres

    @root_validator(pre=True)
    def handle_allocation(cls, values: dict) -> dict:
        if "allocation" in values:
            values["account"] = values.pop("allocation")
        return values

    @root_validator
    def handle_nodes_and_tasks(cls, values: dict) -> dict:
        if (
            values["nodes"] is None
            and values["ntasks"] is None
            and values["ntasks_per_node"] is None
        ):
            values["nodes"] = 1
        return values


class FakeHpcConfig(JadeBaseModel):
    """Defines config options for the fake queueing system."""

    # Keep this required so that Pydantic can differentiate the models.
    walltime: str = Field(
        title="walltime",
        description="Maximum time allocated to each node",
    )


class LocalHpcConfig(JadeBaseModel):
    """Defines config options when there is no HPC."""


class HpcConfig(JadeBaseModel):
    """Defines config options for the HPC."""

    hpc_type: HpcType = Field(
        title="hpc_type",
        description="Type of HPC queueing system (such as 'slurm')",
    )
    job_prefix: str = Field(
        title="job_prefix",
        description="Prefix added to each HPC job name",
        default="job",
    )
    hpc: Union[SlurmConfig, FakeHpcConfig, LocalHpcConfig] = Field(
        title="hpc",
        description="Interface-specific config options",
    )

    @validator("hpc", pre=True)
    def assign_hpc(cls, value, values):
        if isinstance(value, JadeBaseModel):
            return value

        if values["hpc_type"] == HpcType.SLURM:
            return SlurmConfig(**value)
        elif values["hpc_type"] == HpcType.FAKE:
            return FakeHpcConfig(**value)
        elif values["hpc_type"] == HpcType.LOCAL:
            return LocalHpcConfig()
        raise ValueError(f"Unsupported: {values['hpc_type']}")

    def get_num_gpus(self):
        """Return the number of GPUs specified by the config.

        Returns
        -------
        int

        """
        if isinstance(self.hpc, SlurmConfig) and self.hpc.gres is not None:
            return int(self.hpc.gres.split(":")[1])
        return 0



"""HPC management implementation functionality"""

import abc
import getpass
import os

from jade.exceptions import InvalidParameter
from jade.hpc.common import HpcJobStats
from jade.utils.utils import load_data


class HpcManagerInterface(abc.ABC):
    """Defines the implementation interface for managing an HPC."""

    USER = getpass.getuser()

    @abc.abstractmethod
    def am_i_manager(self):
        """Return True if the current node is the manager node.

        Returns
        -------
        bool

        """

    @abc.abstractmethod
    def cancel_job(self, job_id):
        """Cancel job.

        Parameters
        ----------
        job_id : str

        Returns
        -------
        int
            return code

        """

    @abc.abstractmethod
    def check_status(self, name=None, job_id=None):
        """Check the status of a job. Either name or job_id must be passed.
        Handles transient errors for up to one minute.

        Parameters
        ----------
        name : str
            job name
        job_id : str
            job ID

        Returns
        -------
        HpcJobInfo

        Raises
        ------
        ExecutionError
            Raised if statuses cannot be retrieved.

        """

    @abc.abstractmethod
    def check_statuses(self):
        """Check the statuses of all user jobs.
        Handles transient errors for up to one minute.

        Returns
        -------
        dict
            key is job_id, value is HpcJobStatus

        Raises
        ------
        ExecutionError
            Raised if statuses cannot be retrieved.

        """

    @abc.abstractmethod
    def check_storage_configuration(self):
        """Checks if the storage configuration is appropriate for execution.

        Raises
        ------
        InvalidConfiguration
            Raised if the configuration is not valid

        """

    @abc.abstractmethod
    def create_submission_script(self, name, script, filename, path):
        """Create the script to queue the jobs to the HPC.

        Parameters
        ----------
        name : str
            job name
        script : str
            script to execute on HPC
        filename : str
            submission script filename
        path : str
            path for stdout and stderr files

        """

    @abc.abstractmethod
    def create_cluster(self):
        """Create a Dask cluster.

        Returns
        -------
        Dask cluster
            SLURM: SLURMCluster

        """

    @abc.abstractmethod
    def create_local_cluster(self):
        """Create a Dask local cluster.

        Returns
        -------
        dask.distributed.LocalCluster

        """

    @abc.abstractmethod
    def get_config(self):
        """Get HPC configuration parameters.

        Returns
        -------
        dict

        """

    @abc.abstractmethod
    def get_current_job_id(self):
        """Get the job ID for the local compute node.

        Returns
        -------
        str

        """

    @abc.abstractmethod
    def get_job_stats(self, job_id):
        """Get stats for job ID.

        Returns
        -------
        HpcJobStats

        """

    @abc.abstractmethod
    def get_local_scratch(self):
        """Get path to local storage space.

        Returns
        -------
        str

        """

    @abc.abstractmethod
    def get_node_id(self):
        """Return the node ID of the current system.

        Returns
        -------
        str

        """

    @abc.abstractmethod
    def get_num_cpus(self):
        """Return the number of CPUs in the system.

        Returns
        -------
        int

        """

    @abc.abstractmethod
    def list_active_nodes(self, job_id):
        """Return the nodes currently participating in the job. Order should be deterministic.

        Parameters
        ----------
        job_id : str

        Returns
        -------
        list
            list of node hostnames

        """

    @abc.abstractmethod
    def log_environment_variables(self):
        """Logs all relevant HPC environment variables."""

    @abc.abstractmethod
    def submit(self, filename):
        """Submit the work to the HPC queue.
        Handles transient errors for up to one minute.

        Parameters
        ----------
        filename : str
            HPC script filename

        Returns
        -------
        tuple of Status, str, str
            (Status, job_id, stderr)

        """



"""CLI to resubmit failed and missing jobs."""

import logging
import os
import sys
from pathlib import Path

import click

from jade.common import CONFIG_FILE, EVENTS_DIR
from jade.enums import Status
from jade.jobs.cluster import Cluster
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.jobs.results_aggregator import ResultsAggregator
from jade.jobs.job_submitter import JobSubmitter
from jade.loggers import setup_logging, setup_event_logging
from jade.models.submission_group import SubmissionGroup
from jade.result import ResultsSummary
from jade.utils.utils import load_data


logger = logging.getLogger(__name__)


@click.command()
@click.argument(
    "output",
    type=click.Path(exists=True),
)
@click.option(
    "--failed/--no-failed",
    is_flag=True,
    default=True,
    show_default=True,
    help="Resubmit failed and canceled jobs.",
)
@click.option(
    "--missing/--no-missing",
    is_flag=True,
    default=True,
    show_default=True,
    help="Resubmit missing jobs.",
)
@click.option(
    "--successful/--no-successful",
    is_flag=True,
    default=False,
    show_default=True,
    help="Resubmit successful jobs.",
)
@click.option(
    "-s",
    "--submission-groups-file",
    default=None,
    type=click.Path(exists=True),
    help="Use submission parameters in this file generated from 'jade config '"
    "save-submission-groups'",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def resubmit_jobs(output, failed, missing, successful, submission_groups_file, verbose):
    """Resubmit jobs."""
    event_file = os.path.join(output, "submit_jobs_events.log")
    setup_event_logging(event_file, mode="a")
    filename = os.path.join(output, "submit_jobs.log")
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, filename, file_level=level, console_level=level, mode="a")

    cluster, promoted = Cluster.deserialize(
        output,
        try_promote_to_submitter=True,
        deserialize_jobs=True,
    )
    if not cluster.is_complete():
        cluster.demote_from_submitter()
        print("resubmit-jobs requires that the existing submission be complete", file=sys.stderr)
        sys.exit(1)
    assert promoted

    if submission_groups_file is not None:
        groups = load_data(submission_groups_file)
        cur = len(groups)
        orig = len(cluster.config.submission_groups)
        if cur != orig:
            print(
                f"Length of submission_groups ({cur}) must be identical to the original ({orig})",
                file=sys.stderr,
            )
            cluster.demote_from_submitter()
            sys.exit(1)

        for _group in groups:
            group = SubmissionGroup(**_group)
            found = False
            for i, orig_group in enumerate(cluster.config.submission_groups):
                if group.name == orig_group.name:
                    cluster.config.submission_groups[i] = group
                    found = True
                    break
            if not found:
                print(
                    f"submission group {group.name} does not exist in the original",
                    file=sys.stderr,
                )
                cluster.demote_from_submitter()
                sys.exit(1)
        logger.info("Updated submitter parameters from %s", submission_groups_file)

    jobs_to_resubmit = _get_jobs_to_resubmit(cluster, output, failed, missing, successful)
    updated_blocking_jobs_by_name = _update_with_blocking_jobs(jobs_to_resubmit, output)
    _reset_results(output, jobs_to_resubmit)
    cluster.prepare_for_resubmission(jobs_to_resubmit, updated_blocking_jobs_by_name)
    events_dir = Path(output) / EVENTS_DIR
    for path in list(events_dir.iterdir()):
        # These files will get regenerated. It would be better to only generate events for new
        # compute node batches, but the code in events.py doesn't support that.
        # TODO
        path.unlink()

    ret = 1
    try:
        mgr = JobSubmitter.load(output)
        status = mgr.submit_jobs(cluster)
        if status == Status.IN_PROGRESS:
            print(f"Resubmitted {len(jobs_to_resubmit)} jobs in {output}")
            ret = 0
        else:
            ret = status.value
    except Exception:
        logger.exception("Failed to resubmit jobs")
        raise
    finally:
        cluster.demote_from_submitter()

    sys.exit(ret)


def _get_jobs_to_resubmit(cluster, output, failed, missing, successful):
    results = ResultsSummary(output)
    jobs_to_resubmit = []
    if failed or successful:
        res = results.get_results_by_type()
        if failed:
            jobs_to_resubmit += res["canceled"]
            jobs_to_resubmit += res["failed"]
        if successful:
            jobs_to_resubmit += res["successful"]
    if missing:
        jobs_to_resubmit += results.get_missing_jobs(cluster.iter_jobs())

    return {x.name for x in jobs_to_resubmit}


def _update_with_blocking_jobs(jobs_to_resubmit, output):
    config = create_config_from_file(Path(output) / CONFIG_FILE)

    # Any job that was blocked by any of these jobs must also be resubmitted.
    # Same for any job blocked by one of those.
    # Account for abnormal ordering where a lower-ID'd job is blocked by a later one.
    updated_blocking_jobs_by_name = {}
    max_iter = config.get_num_jobs()
    for i in range(max_iter):
        first = len(jobs_to_resubmit)
        for job in config.iter_jobs():
            blocking_jobs = job.get_blocking_jobs()
            if not blocking_jobs:
                continue
            intersecting_jobs = blocking_jobs.intersection(jobs_to_resubmit)
            if intersecting_jobs:
                updated_blocking_jobs_by_name[job.name] = intersecting_jobs
                jobs_to_resubmit.add(job.name)
        num_added = len(jobs_to_resubmit) - first
        if num_added == 0:
            break
        assert i < max_iter - 1, f"max_iter={max_iter} num_added={num_added} first={first}"

    return updated_blocking_jobs_by_name


def _reset_results(output, jobs_to_resubmit):
    aggregator = ResultsAggregator.load(output)
    aggregator.clear_results_for_resubmission(jobs_to_resubmit)



"""This is a batch post-process script.

Merge predicted GDP of all countries into one CSV file.
"""
import logging
import os

import click
import pandas as pd


logger = logging.getLogger(__name__)


@click.group()
def cli():
    pass


@cli.command()
@click.argument("previous_stage_output")
@click.argument("output")
def run(previous_stage_output, output):
    """Merge the pred_gdp in CSV result into one CSV file."""
    logger.info("Start batch post-processing...")

    # Get the job-outputs directory
    job_outputs = os.path.join(previous_stage_output, "job-outputs")

    # Merge pred_gdp columns
    results = pd.DataFrame()
    for job_name in os.listdir(job_outputs):
        csv_result = os.path.join(job_outputs, job_name, "result.csv")
        if not os.path.exists(csv_result):
            continue
        df = pd.read_csv(csv_result)
        if results.empty:
            results["year"] = df["year"]
        results[job_name] = df["pred_gdp"]

    # Export merged results to CSV file
    csv_file = os.path.join(output, "pred_gdp.csv")
    results.to_csv(csv_file, index=False)

    logger.info("Batch post-processing finished!")


if __name__ == "__main__":
    cli()



"""
Demo cli
"""
import os
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.extensions.demo.autoregression_configuration import AutoRegressionConfiguration
from jade.extensions.demo.autoregression_inputs import AutoRegressionInputs
from jade.extensions.demo.autoregression_execution import AutoRegressionExecution


def auto_config(inputs, **kwargs):
    """
    Create a configuration file for demo

    :param inputs: str, the path to directory containing autoregression data.

    :return: None
    """
    if not os.path.exists(inputs):
        raise OSError(f"Inputs path '{inputs}' does not exist.")

    job_inputs = AutoRegressionInputs(inputs)
    config = AutoRegressionConfiguration(**kwargs)
    for job_param in job_inputs.iter_jobs():
        config.add_job(job_param)

    return config


def run(config_file, name, output, output_format, verbose):
    """Run auto regression analysis through command line"""
    os.makedirs(output, exist_ok=True)

    config = create_config_from_file(config_file)
    job = config.get_job(name)

    execution = AutoRegressionExecution(job=job, output=output, output_format=output_format)
    execution.run()



"""Controls submission of jobs to HPC nodes."""

import copy
import itertools
import logging
import os
import sys
import time
from datetime import timedelta
from pathlib import Path

from jade.enums import JobCompletionStatus, Status
from jade.events import (
    StructuredLogEvent,
    EVENT_CATEGORY_HPC,
    EVENT_NAME_HPC_SUBMIT,
    EVENT_NAME_HPC_JOB_ASSIGNED,
    EVENT_NAME_HPC_JOB_STATE_CHANGE,
)
from jade.exceptions import InvalidConfiguration
from jade.hpc.common import HpcJobStatus, HpcType
from jade.hpc.hpc_manager import HpcManager
from jade.jobs.async_job_interface import AsyncJobInterface
from jade.jobs.cluster import Cluster
from jade.jobs.job_configuration import JobConfiguration
from jade.jobs.job_queue import JobQueue
from jade.jobs.results_aggregator import ResultsAggregator
from jade.loggers import log_event
from jade.models import JobState
from jade.models.submission_group import make_submission_group_lookup
from jade.result import Result
from jade.utils.timing_utils import timed_debug
from jade.utils.utils import (
    aggregate_data_from_files,
    dump_data,
    create_script,
    ExtendedJSONEncoder,
)

logger = logging.getLogger(__name__)


class HpcSubmitter:
    """Submits batches of jobs to HPC. Manages job ordering."""

    LOCK_FILENAME = "submitter.lock"

    def __init__(self, config: JobConfiguration, config_file, cluster: Cluster, output):
        self._config = config
        self._submission_groups = make_submission_group_lookup(cluster.config.submission_groups)
        self._config_file = config_file
        self._base_config = config.serialize()
        self._batch_index = cluster.job_status.batch_index
        self._cluster = cluster
        self._hpc_mgr = HpcManager(self._submission_groups, output)
        self._output = output

        # Limitation: these settings apply to all groups in aggregate.
        # This could be made more flexible if needed.
        group = next(iter(self._submission_groups.values()))
        self._max_nodes = group.submitter_params.max_nodes
        if self._max_nodes is None:
            self._max_nodes = sys.maxsize
        self._poll_interval = group.submitter_params.poll_interval
        self._status_collector = HpcStatusCollector(self._hpc_mgr, self._poll_interval)

    def _create_run_script(self, config_file, filename, submission_group):
        text = ["#!/bin/bash"]
        sing_params = submission_group.submitter_params.singularity_params
        if sing_params and sing_params.enabled:
            text += sing_params.setup_commands.split("\n")
        if submission_group.submitter_params.distributed_submitter:
            dsub = "--distributed-submitter"
        else:
            dsub = "--no-distributed-submitter"
        command = f"jade-internal run-jobs {config_file} --output={self._output} {dsub}"
        if submission_group.submitter_params.num_parallel_processes_per_node is not None:
            command += f" --num-parallel-processes-per-node={submission_group.submitter_params.num_parallel_processes_per_node}"
        if submission_group.submitter_params.verbose:
            command += " --verbose"

        text.append(command)
        create_script(filename, "\n".join(text) + "\n")

    def _make_async_submitter(self, jobs, submission_group, dry_run=False):
        config = copy.copy(self._base_config)
        config["jobs"] = jobs
        suffix = f"_batch_{self._batch_index}"
        self._batch_index += 1
        new_config_file = str(self._config_file).replace(".json", f"{suffix}.json")
        dump_data(config, new_config_file, cls=ExtendedJSONEncoder)
        logger.info(
            "Created split config file %s with %s jobs", new_config_file, len(config["jobs"])
        )

        run_script = os.path.join(self._output, f"run{suffix}.sh")
        self._create_run_script(new_config_file, run_script, submission_group)

        name = submission_group.submitter_params.hpc_config.job_prefix + suffix
        return AsyncHpcSubmitter(
            self._hpc_mgr,
            self._status_collector,
            run_script,
            name,
            submission_group,
            self._output,
            dry_run=dry_run,
        )

    @timed_debug
    def run(self):
        """Try to submit batches of jobs to the HPC.

        Returns
        -------
        bool
            Returns True if all jobs are complete.

        """
        starting_batch_index = self._batch_index
        # TODO: consider whether we need to save the real job names
        hpc_submitters = [
            AsyncHpcSubmitter.create_from_id(self._hpc_mgr, self._status_collector, x)
            for x in self._cluster.iter_hpc_job_ids()
        ]

        queue = JobQueue(
            self._max_nodes,
            existing_jobs=hpc_submitters,
            poll_interval=self._poll_interval,
        )
        # Statuses may have changed since we last ran.
        # Persistent network errors could cause this submitter to fail.
        # Another submitter will try again later (unless this is the last submitter).
        queue.process_queue()
        completed_job_names, canceled_jobs = self._update_completed_jobs()

        lock_file = Path(self._output) / self.LOCK_FILENAME
        if lock_file.exists():
            raise Exception(
                f"{lock_file} exists. A previous submitter crashed in an unknown state."
            )
        lock_file.touch()

        # Start submitting jobs. If any unexpected exception prevents us from updating the
        # status file, leave the lock_file in place and intentionally cause a deadlock.
        try:
            blocked_jobs = []
            submitted_jobs = []
            for group in self._cluster.config.submission_groups:
                if not queue.is_full():
                    self._submit_batches(queue, group, blocked_jobs, submitted_jobs)

            num_submissions = self._batch_index - starting_batch_index
            logger.info(
                "num_batches=%s num_submitted=%s num_blocked=%s new_completions=%s",
                num_submissions,
                len(submitted_jobs),
                len(blocked_jobs),
                len(completed_job_names),
            )

            hpc_job_ids = sorted([x.job_id for x in queue.outstanding_jobs])
            self._update_status(
                submitted_jobs,
                blocked_jobs,
                canceled_jobs,
                hpc_job_ids,
                completed_job_names,
            )

            is_complete = self._is_complete()
            os.remove(lock_file)
            return is_complete
        except Exception:
            logger.exception(
                "An exception occurred while the submitter was active. "
                "The state of the cluster is unknown. A deadlock will occur."
            )
            raise

    def _update_status(
        self, submitted_jobs, blocked_jobs, canceled_jobs, hpc_job_ids, completed_job_names
    ):
        hpc_job_changes = self._cluster.job_status.hpc_job_ids != hpc_job_ids
        if completed_job_names or submitted_jobs or blocked_jobs or hpc_job_changes:
            self._cluster.update_job_status(
                submitted_jobs,
                blocked_jobs,
                canceled_jobs,
                completed_job_names,
                hpc_job_ids,
                self._batch_index,
            )

    def _submit_batches(self, queue, submission_group, blocked_jobs, submitted_jobs):
        assert not queue.is_full()
        num_submitted_jobs = 0
        _submitted_jobs = []  # only exists for the check at the end
        if (
            not submission_group.submitter_params.time_based_batching
            or "JADE_SKIP_SORT_BY_TIME" in os.environ
        ):
            available_jobs = self._get_available_jobs(submission_group)
        else:
            available_jobs = self._get_available_jobs_by_time(submission_group)
        while not queue.is_full() and available_jobs:
            batch, available_jobs = self._make_batch(
                available_jobs, submission_group, _submitted_jobs, blocked_jobs
            )
            if batch.num_jobs > 0:
                self._submit_batch(queue, submission_group, batch)
                num_submitted_jobs += batch.num_jobs

        assert num_submitted_jobs == len(
            _submitted_jobs
        ), f"{num_submitted_jobs} / {len(_submitted_jobs)}"
        submitted_jobs.extend(_submitted_jobs)

    def _get_available_jobs(self, submission_group):
        available_jobs = []
        for job in self._cluster.iter_jobs(state=JobState.NOT_SUBMITTED):
            jade_job = self._config.get_job(job.name)
            if jade_job.submission_group == submission_group.name:
                available_jobs.append(job)
        return available_jobs

    def _get_available_jobs_by_time(self, submission_group):
        available_jobs = {}
        job_order = []
        for job in self._cluster.iter_jobs(state=JobState.NOT_SUBMITTED):
            jade_job = self._config.get_job(job.name)
            if jade_job.submission_group == submission_group.name:
                job_order.append((job.name, jade_job.estimated_run_minutes))
                available_jobs[job.name] = job
        job_order.sort(key=lambda x: x[1])

        logger.info("Sorted jobs by estimated_run_minutes.")
        return [available_jobs[x[0]] for x in job_order]

    def _make_batch(self, available_jobs, submission_group, submitted_jobs, blocked_jobs):
        blocked_jobs_by_name = {}
        submitted_jobs_by_name = set()
        batch = _BatchJobs(submission_group.submitter_params)
        if submission_group.submitter_params.try_add_blocked_jobs:
            # Allow multiple rounds in case the user listed blocked jobs before
            # their blocking jobs in the config file.
            max_iterations = len(available_jobs)
        else:
            max_iterations = 1
        highest_index = -1
        done = False
        for _ in range(max_iterations):
            for i, job in enumerate(available_jobs):
                if i > highest_index:
                    highest_index = i
                if job.name in submitted_jobs_by_name:
                    continue
                jade_job = self._config.get_job(job.name)
                if batch.is_job_blocked(job):
                    blocked_jobs_by_name[job.name] = job
                else:
                    jade_job.set_blocking_jobs(job.blocked_by)
                    if batch.try_append(jade_job):
                        submitted_jobs.append(job)
                        submitted_jobs_by_name.add(job.name)
                        blocked_jobs_by_name.pop(job.name, None)
                    else:
                        # Need to look at this job in the next round.
                        highest_index -= 1
                if batch.is_ready_to_submit or len(submitted_jobs_by_name) == len(available_jobs):
                    done = True
                    break
            if done:
                break

        for job in blocked_jobs_by_name.values():
            blocked_jobs.append(job)
        if highest_index == len(available_jobs) - 1:
            not_checked = []
        else:
            not_checked = available_jobs[highest_index + 1 :]
        return batch, not_checked

    def _submit_batch(self, queue, submission_group, batch):
        async_submitter = self._make_async_submitter(
            batch.serialize(), submission_group, dry_run=submission_group.submitter_params.dry_run
        )
        queue.submit(async_submitter)
        self._log_submission_event(submission_group, batch)

    def _is_complete(self):
        is_complete = self._cluster.are_all_jobs_complete()
        if not is_complete and not self._cluster.job_status.hpc_job_ids:
            # TODO: need to implement persistent recording of fake status
            if self._hpc_mgr.hpc_type != HpcType.FAKE:
                logger.error(
                    "Some jobs are not complete but there are no active HPC job IDs. "
                    "Force completion."
                )
                is_complete = True

        return is_complete

    def _log_submission_event(self, submission_group, batch):
        event = StructuredLogEvent(
            source=submission_group.submitter_params.hpc_config.job_prefix,
            category=EVENT_CATEGORY_HPC,
            name=EVENT_NAME_HPC_SUBMIT,
            message="Submitted HPC batch",
            batch_size=batch.num_jobs,
            per_node_batch_size=submission_group.submitter_params.per_node_batch_size,
        )
        log_event(event)

    def _cancel_job(self, job, aggregator):
        job.state = JobState.DONE
        job.blocked_by.clear()
        result = Result(job.name, 1, JobCompletionStatus.CANCELED, 0, hpc_job_id=None)
        aggregator.append_result(result)
        logger.info("Canceled job %s because one of its blocking jobs failed.", job.name)
        return result

    def _update_completed_jobs(self):
        newly_completed = set()
        canceled_jobs = []
        # If jobs fail and are configured to cancel blocked jobs, we may need to run this
        # loop many times to cancel the entire chain.
        aggregator = ResultsAggregator.load(self._output)
        need_to_rerun = True
        new_results = []
        while need_to_rerun:
            need_to_rerun = False
            failed_jobs = set()
            for result in itertools.chain(aggregator.process_results(), new_results):
                newly_completed.add(result.name)
                if result.return_code != 0:
                    failed_jobs.add(result.name)
            new_results.clear()

            logger.debug("Detected completion of jobs: %s", newly_completed)
            logger.debug("Detected failed jobs: %s", failed_jobs)
            for job in self._cluster.iter_jobs(state=JobState.NOT_SUBMITTED):
                if job.blocked_by:
                    if job.cancel_on_blocking_job_failure and job.blocked_by.intersection(
                        failed_jobs
                    ):
                        result = self._cancel_job(job, aggregator)
                        canceled_jobs.append(job)
                        new_results.append(result)
                        need_to_rerun = True
                    else:
                        job.blocked_by.difference_update(newly_completed)

        return newly_completed, canceled_jobs


class _BatchJobs:
    """Helper class to manage jobs in a batch."""

    def __init__(self, params):
        self._estimated_batch_time = timedelta(seconds=0)
        self._num_processes = params.num_parallel_processes_per_node
        self._per_node_batch_size = params.per_node_batch_size
        self._time_based_batching = params.time_based_batching
        self._try_add_blocked_jobs = params.try_add_blocked_jobs
        self._jobs = []
        self._job_names = set()
        self._is_ready_to_submit = False
        if self._time_based_batching:
            self._max_batch_time = params.get_wall_time() * self._num_processes
        else:
            self._max_batch_time = None

    def try_append(self, job):
        """Return True if the job is appended to the batch.

        Returns
        -------
        bool

        """
        # This is probably too simplistic and could be made more robust.
        # It does assume that if time_based_batching is enabled, jobs are sorted by
        # estimated_run_minutes in ascending order. Once one job is too long, all other
        # jobs will be too long.
        if (
            self._time_based_batching
            and self._estimated_batch_time + timedelta(minutes=job.estimated_run_minutes)
            > self._max_batch_time
        ):
            self._is_ready_to_submit = True
            return False

        self._jobs.append(job)
        self._job_names.add(job.name)

        if self._time_based_batching:
            self._estimated_batch_time += timedelta(minutes=job.estimated_run_minutes)
        elif self.num_jobs >= self._per_node_batch_size:
            self._is_ready_to_submit = True
        return True

    def are_blocking_jobs_present(self, blocking_jobs):
        """Return True if all blocking jobs are already in the batch.

        Returns
        -------
        bool

        """
        return blocking_jobs.issubset(self._job_names)

    def is_job_blocked(self, job):
        """Return True if the job is blocked.

        Parameters
        ----------
        job : Job

        Returns
        -------
        bool

        """
        if not job.blocked_by:
            return False
        if self._try_add_blocked_jobs and self.are_blocking_jobs_present(job.blocked_by):
            # JobRunner will manage the execution ordering on the compute node.
            return False
        return True

    @property
    def is_ready_to_submit(self):
        """Return True if the batch has enough jobs to submit."""
        return self._is_ready_to_submit

    @property
    def num_jobs(self):
        """Return the number of jobs in the batch."""
        return len(self._jobs)

    def serialize(self):
        """Serialize all jobs in the batch.

        Returns
        -------
        list
            list of dict

        """
        return [x.serialize() for x in self._jobs]


class AsyncHpcSubmitter(AsyncJobInterface):
    """Used to submit batches of jobs to multiple nodes, one at a time."""

    def __init__(
        self,
        hpc_manager,
        status_collector,
        run_script,
        name,
        submission_group,
        output,
        job_id=None,
        dry_run=False,
    ):
        self._mgr = hpc_manager
        self._status_collector = status_collector
        self._run_script = Path(run_script) if run_script else None
        self._submission_group = submission_group
        self._job_id = job_id
        self._output = output
        self._name = name
        # Re-enable this if we ever keep a submitter active and polling again.
        # self._last_status = HpcJobStatus.NONE
        self._is_complete = False
        self._dry_run = dry_run
        self._return_code = None

    def cancel(self):
        self._mgr.cancel_job(self._job_id)

    @property
    def cancel_on_blocking_job_failure(self):
        return False

    @classmethod
    def create_from_id(cls, hpc_manager, status_collector, job_id):
        """Create an instance of a job_id in order to check status.

        Parameters
        ----------
        job_id : str

        """
        return cls(hpc_manager, status_collector, None, job_id, None, None, job_id=job_id)

    @property
    def hpc_manager(self):
        """Return the HpcManager object.

        Returns
        -------
        HpcManager

        """
        return self._mgr

    def is_complete(self):
        if self._is_complete:
            return self._is_complete
        status = self._status_collector.check_status(self._job_id)
        # if status != self._last_status:
        logger.info("HPC job ID %s status=%s", self._job_id, status)
        #    event = StructuredLogEvent(
        #        source=self._name,
        #        category=EVENT_CATEGORY_HPC,
        #        name=EVENT_NAME_HPC_JOB_STATE_CHANGE,
        #        message="HPC job state change",
        #        job_id=self._job_id,
        #        old_state=self._last_status.value,
        #        new_state=status.value,
        #    )
        #    log_event(event)
        #    self._last_status = status

        self._is_complete = status in (HpcJobStatus.COMPLETE, HpcJobStatus.NONE)
        return self._is_complete

    def get_id(self):
        return self.job_id

    @property
    def job_id(self):
        return self._job_id

    @property
    def name(self):
        return self._name

    @property
    def return_code(self):
        assert self._is_complete
        return self._return_code

    def run(self):
        assert self._submission_group is not None
        sing_params = self._submission_group.submitter_params.singularity_params
        if sing_params and sing_params.enabled:
            script = self._make_singularity_command()
        else:
            script = self._run_script
        job_id, result = self._mgr.submit(
            self._output,
            self._name,
            str(script),
            self._submission_group.name,
            dry_run=self._dry_run,
        )
        if result != Status.GOOD:
            # TODO: cancel or fail all jobs in the batch
            logger.error("Failed to submit name=%s", self._name)
            self._return_code = 1
            self._is_complete = True
            return Status.ERROR

        self._job_id = job_id
        event = StructuredLogEvent(
            source=self._name,
            category=EVENT_CATEGORY_HPC,
            name=EVENT_NAME_HPC_JOB_ASSIGNED,
            message="HPC job assigned",
            job_id=self._job_id,
        )
        log_event(event)
        logger.info("Assigned job_ID=%s name=%s", self._job_id, self._name)
        return Status.GOOD

    def get_blocking_jobs(self):
        return set()

    def remove_blocking_job(self, name):
        assert False

    def set_blocking_jobs(self, jobs):
        assert False

    def _make_singularity_command(self):
        params = self._submission_group.submitter_params.singularity_params
        if not params.run_command:
            raise InvalidConfiguration("Singularity command cannot be empty.")
        container_path = Path(params.container)
        cmd = f"{params.run_command} {container_path} {self._run_script}"
        sing_script = self._run_script.parent / self._run_script.name.replace(
            "run", "singularity", 1
        )
        text = "#!/bin/bash\n" + params.load_command + "\n" + cmd + "\n"
        create_script(str(sing_script), text)
        return sing_script


class HpcStatusCollector:
    """Collects status for all user jobs."""

    def __init__(self, hpc_mgr, poll_interval):
        self._hpc_mgr = hpc_mgr
        self._poll_interval = poll_interval
        self._last_poll_time = None
        self._statuses = {}

    def check_status(self, job_id):
        """Return the status for job_id.

        Parameters
        ----------
        job_id : str

        Returns
        -------
        HpcJobStatus

        Raises
        ------
        ExecutionError
            Raised if statuses cannot be retrieved.

        """
        cur_time = time.time()
        if self._last_poll_time is None or cur_time - self._last_poll_time > self._poll_interval:
            logger.debug("Collect new statuses.")
            self._statuses = self._hpc_mgr.check_statuses()
            self._last_poll_time = cur_time

        return self._statuses.get(job_id, HpcJobStatus.NONE)

    def get_statuses(self):
        """Return outstanding statuses

        Returns
        -------
        list
            list of HpcJobStatus

        """
        return list(self._statuses.values())



"""Functionality to execute non-blocking commands in a subprocess."""

import logging
import shlex
import subprocess
import sys
import threading
import time

from jade.exceptions import JobAlreadyInProgress

# These functions used to be in this file. Leave this here for backwards compatibility.
from .run_command import run_command, check_run_command


logger = logging.getLogger(__name__)


class SubprocessManager:
    """Allows non-blocking execution of a command in a subprocess."""

    def __init__(self):
        self._pipe = None
        self._return_code = None
        self._shutdown = False
        self._thread = None

    def _run_worker(self, cmd, timeout):
        if timeout is None:
            # Effectively, no timeout.
            timeout = sys.maxsize

        self._return_code = None
        start = time.time()
        self._pipe = subprocess.Popen(shlex.split(cmd))

        ret = None
        elapsed_time = 0
        sleep_interval = 1

        while not self._shutdown and ret is None and elapsed_time < timeout:
            time.sleep(sleep_interval)
            elapsed_time += sleep_interval
            ret = self._pipe.poll()

        exec_time = time.time() - start
        self._return_code = ret

        if self._shutdown:
            logger.info("Exit job thread for shutdown")
            self._terminate()
        elif self._return_code is None:
            logger.error("Command [%s] timed out, terminate it; timeout=%s", cmd, timeout)
            self._terminate()
        else:
            logger.info(
                "Command [%s] completed return_code=%s " "exec_time_s=%s",
                cmd,
                self._return_code,
                exec_time,
            )

        self._pipe = None

    def _terminate(self):
        """Should only be called by execution thread."""
        assert threading.get_ident() == self._thread.ident

        if self._pipe is not None:
            self._pipe.terminate()
        # else it may have been terminated before actually starting the cmd.

    def in_progress(self):
        """Returns True if commands are still running."""
        return self._thread is not None and self._thread.is_alive()

    @property
    def return_code(self):
        """Return the return coe of the last completed command.

        Returns
        -------
        int | None
            Returns None if the command did not complete.

        """
        return self._return_code

    def run(self, command, timeout=None):
        """Run a command without blocking. Call wait_for_completion() or
        in_progress() to monitor progress.

        Parameters
        ----------
        command : str
            Command to execute
        timeout: int or None
            Timeout in seconds for each command. None means no timeout.

        Raises
        ------
        JobAlreadyInProgress
            Raised if execution is already in progress.

        """
        if self.in_progress():
            raise JobAlreadyInProgress(
                "SubprocessManager only supports " "execution of one job at a time."
            )

        logger.debug("Start command execution thread.")

        self._thread = threading.Thread(
            target=self._run_worker, args=(command, timeout), name="run_command"
        )

        # The thread (and subprocesses) will get killed if the main process
        # is killed.
        self._thread.daemon = True

        self._thread.start()

    def terminate(self):
        """Terminates the command."""
        logger.info("Terminate execution.")
        self._shutdown = True
        self.wait_for_completion()

    def wait_for_completion(self):
        """Waits until the command is complete.

        Returns
        -------
        int | None
            return code if the command completed, otherwise None

        """
        try:
            self._thread.join()
            return self._return_code
        except KeyboardInterrupt:
            logger.info("Detected Ctrl-C: command is still outstanding")
            self._shutdown = True
            return None



"""Synchronizes updates to the results file across jobs."""

import csv
import glob
import logging
import os
import time
from pathlib import Path

from filelock import SoftFileLock, Timeout

from jade.common import RESULTS_DIR
from jade.result import Result, deserialize_result, serialize_result


LOCK_TIMEOUT = 300
PROCESSED_RESULTS_FILENAME = "processed_results.csv"

logger = logging.getLogger(__name__)


class ResultsAggregator:
    """Synchronizes updates to the results file.

    One instance is used to aggregate results from all compute nodes.
    One instance is used for each compute node.
    """

    def __init__(self, filename, timeout=LOCK_TIMEOUT, delimiter=","):
        """
        Constructs ResultsAggregator.

        Parameters
        ----------
        filename : Path
            Results file.
        timeout : int
            Lock acquistion timeout in seconds.
        delimiter : str
            Delimiter to use for CSV formatting.

        """
        self._filename = filename
        self._lock_file = self._filename.parent / (self._filename.name + ".lock")
        self._timeout = timeout
        self._delimiter = delimiter
        self._is_node = "batch" in filename.name

    @classmethod
    def create(cls, output_dir, **kwargs):
        """Create a new instance.

        Parameters
        ----------
        output_dir : str

        Returns
        -------
        ResultsAggregator

        """
        agg = cls(Path(output_dir) / PROCESSED_RESULTS_FILENAME, **kwargs)
        agg.create_files()
        return agg

    @classmethod
    def load(cls, output_dir, **kwargs):
        """Load an instance from an output directory.

        Parameters
        ----------
        output_dir : str

        Returns
        -------
        ResultsAggregator

        """
        return cls(Path(output_dir) / PROCESSED_RESULTS_FILENAME, **kwargs)

    @classmethod
    def load_node_results(cls, output_dir, batch_id, **kwargs):
        """Load a per-node instance from an output directory.

        Parameters
        ----------
        output_dir : str
        batch_id : int

        Returns
        -------
        ResultsAggregator

        """
        return cls(Path(output_dir) / RESULTS_DIR / f"results_batch_{batch_id}.csv", **kwargs)

    @classmethod
    def load_node_results_file(cls, path, **kwargs):
        """Load a per-node instance from an output directory.

        Parameters
        ----------
        path : Path

        Returns
        -------
        ResultsAggregator

        """
        return cls(path, **kwargs)

    @staticmethod
    def _get_fields():
        return Result._fields

    def _do_action_under_lock(self, func, *args, **kwargs):
        # Using this instead of FileLock because it will be used across nodes
        # on the Lustre filesystem.
        lock = SoftFileLock(self._lock_file, timeout=self._timeout)
        start = time.time()
        try:
            lock.acquire(timeout=self._timeout)
        except Timeout:
            # Picked a default value such that this should not trip. If it does
            # trip under normal circumstances then we need to reconsider this.
            logger.error(
                "Failed to acquire file lock %s within %s seconds", self._lock_file, self._timeout
            )
            raise

        duration = time.time() - start
        if duration > 10:
            logger.warning("Acquiring ResultsAggregator lock took too long: %s", duration)

        try:
            return func(*args, **kwargs)
        finally:
            lock.release()

    def create_files(self):
        """Initialize the results file. Should only be called by the parent
        process.

        """
        self._do_action_under_lock(self._create_files)

    def _create_files(self):
        with open(self._filename, "w") as f_out:
            f_out.write(self._delimiter.join(self._get_fields()))
            f_out.write("\n")

    @classmethod
    def append(cls, output_dir, result, batch_id=None):
        """Append a result to the file.

        output_dir : str
        result : Result
        batch_id : int

        """
        if batch_id is None:
            aggregator = cls.load(output_dir)
        else:
            aggregator = cls.load_node_results(output_dir, batch_id)
        aggregator.append_result(result)

    def append_result(self, result):
        """Append a result to the file.

        result : Result

        """
        start = time.time()
        text = self._delimiter.join([str(getattr(result, x)) for x in self._get_fields()])
        self._do_action_under_lock(self._append_result, text)
        duration = time.time() - start
        if duration > 10:
            logger.warning("Appending a result took too long: %s", duration)

    def _append_result(self, text):
        with open(self._filename, "a") as f_out:
            if f_out.tell() == 0:
                f_out.write(self._delimiter.join(self._get_fields()))
                f_out.write("\n")
            f_out.write(text)
            f_out.write("\n")

    def _append_processed_results(self, results):
        assert not self._is_node
        with open(self._filename, "a") as f_out:
            for result in results:
                text = self._delimiter.join([str(getattr(result, x)) for x in self._get_fields()])
                f_out.write(text)
                f_out.write("\n")

    def clear_results_for_resubmission(self, jobs_to_resubmit):
        """Remove jobs that will be resubmitted from the results file.

        Parameters
        ----------
        jobs_to_resubmit : set
            Job names that will be resubmitted.

        """
        results = [x for x in self.get_results() if x.name not in jobs_to_resubmit]
        self._write_results(results)
        logger.info("Cleared %s results from %s", len(results), self._filename)

    def clear_unsuccessful_results(self):
        """Remove failed and canceled results from the results file."""
        results = [x for x in self.get_results() if x.return_code == 0]
        self._write_results(results)
        logger.info("Cleared failed results from %s", self._filename)

    def _write_results(self, results):
        _results = [serialize_result(x) for x in results]
        with open(self._filename, "w") as f_out:
            writer = csv.DictWriter(f_out, fieldnames=Result._fields)
            writer.writeheader()
            if results:
                writer.writerows(_results)

    def get_results(self):
        """Return the current results.

        Returns
        -------
        list
            list of Result objects

        """
        return self._do_action_under_lock(self._get_all_results)

    def get_results_unsafe(self):
        """Return the results. It is up to the caller to ensure that
        a lock is not needed.

        Returns
        -------
        list
            list of Result objects

        """
        return self._get_results()

    def _get_all_results(self):
        unprocessed_results = list((self._filename / RESULTS_DIR).glob("results*.csv"))
        if unprocessed_results:
            logger.error("Found unprocessed results: %s", unprocessed_results)
        # TODO: Older code included unprocessed results here. Not sure why.
        return self._get_results()

    def _get_results(self):
        with open(self._filename) as f_in:
            results = []
            reader = csv.DictReader(f_in, delimiter=self._delimiter)
            for row in reader:
                row["return_code"] = int(row["return_code"])
                row["exec_time_s"] = float(row["exec_time_s"])
                row["completion_time"] = float(row["completion_time"])
                result = deserialize_result(row)
                results.append(result)

            return results

    def move_results(self, func):
        """Move the results to a new location and delete the file.

        Parameters
        ----------
        func : function

        Returns
        -------
        list
            list of Result

        """
        return self._do_action_under_lock(self._move_results, func)

    def _move_results(self, func):
        results = self._get_results()
        func(results)
        os.remove(self._filename)
        return results

    @classmethod
    def list_results(cls, output_dir, **kwargs):
        """Return the current results.

        Parameters
        ----------
        output_dir : str

        Returns
        -------
        list
            list of Result objects

        """
        results = cls.load(output_dir, **kwargs)
        return results.get_results()

    def process_results(self):
        """Move all temp results into the consolidated file, then clear the file.

        Returns
        -------
        list
            list of Result objects that are newly completed

        """
        assert not self._is_node
        return self._do_action_under_lock(self._process_results)

    def _get_node_results_files(self):
        assert not self._is_node
        return list((self._filename.parent / RESULTS_DIR).glob("results_batch_*.csv"))

    def _process_results(self):
        results = []
        for path in self._get_node_results_files():
            agg = ResultsAggregator.load_node_results_file(path)
            results += agg.move_results(self._append_processed_results)
        return results



"""Defines interface for job containers."""

import abc


class JobContainerInterface(abc.ABC):
    """Defines interface for job containers."""

    @abc.abstractmethod
    def __iter__(self):
        pass

    @abc.abstractmethod
    def __len__(self):
        pass

    @abc.abstractmethod
    def add_job(self, job):
        """Add a job to the configuration.

        Parameters
        ----------
        job : JobParametersInterface

        """

    @abc.abstractmethod
    def clear(self):
        """Clear all configured jobs."""

    @abc.abstractmethod
    def get_job(self, name):
        """Return the job matching name.

        Returns
        -------
        namedtuple

        """

    @abc.abstractmethod
    def get_jobs(self, sort=False):
        """Return all jobs.

        Parameters
        ----------
        sort : bool

        Returns
        -------
        list

        """

    def list_jobs(self):
        """Return a list of all jobs.

        Returns
        ------
        list
            list of JobParametersInterface

        """
        return list(iter(self))

    @abc.abstractmethod
    def remove_job(self, job):
        """Remove a job from the configuration.

        Parameters
        ----------
        job : JobParametersInterface

        """

    @abc.abstractmethod
    def shuffle(self):
        """Shuffle the order of the jobs."""



"""Contains pandas dataframe utility functions."""

import gzip
import logging
import os
import re
import shutil

import pandas as pd

from jade.exceptions import InvalidParameter
from jade.utils.timing_utils import timed_debug


logger = logging.getLogger(__name__)


@timed_debug
def read_dataframe(filename, index_col=None, columns=None, parse_dates=False, **kwargs):
    """Convert filename to a dataframe. Supports .csv, .json, .feather, .h5.
    Handles compressed files.

    Parameters
    ----------
    filename : str
    index_col : str | int | None
        Index column name or index
    columns : list or None
        Use these columns if the file is CSV and does not define them.
    parse_dates : bool
    kwargs : kwargs
        Passed to underlying library for dataframe conversion.
        Consider setting parse_dates=True if the index is a timestamp.

    Returns
    -------
    pd.DataFrame

    Raises
    ------
    FileNotFoundError
        Raised if the file does not exist.

    """
    if not os.path.exists(filename):
        raise FileNotFoundError("filename={} does not exist".format(filename))

    needs_new_index = False
    ext = os.path.splitext(filename)
    if ext[1] == ".gz":
        ext = os.path.splitext(ext[0])[1]
        open_func = gzip.open
    else:
        ext = ext[1]
        open_func = open

    if ext == ".csv":
        df = pd.read_csv(
            filename, index_col=index_col, usecols=columns, parse_dates=parse_dates, **kwargs
        )
    elif ext == ".json":
        df = pd.read_json(filename, **kwargs)
    elif ext == ".feather":
        needs_new_index = True
        with open_func(filename, "rb") as f_in:
            df = pd.read_feather(f_in, **kwargs)
    elif ext == ".h5":
        # This assumes that the file has a single dataframe, and so the
        # key name is not relevant.
        df = pd.read_hdf(filename, **kwargs)
        needs_new_index = True
    else:
        raise InvalidParameter(f"unsupported file extension {ext}")

    if index_col is not None and needs_new_index:
        df.set_index(index_col, inplace=True)
        if parse_dates:
            df.set_index(pd.to_datetime(df.index), inplace=True)

    return df


def read_dataframe_handle_missing(filename, index_col=None, columns=None):
    """Convert filename to a dataframe. Returns None if the file is missing.

    Parameters
    ----------
    filename : str
    index_col : str | int | None
        Index column name or index
    columns : list or None
        Use these columns if the file is CSV and does not define them.

    Returns
    -------
    pd.DataFrame | None

    """
    if not os.path.exists(filename):
        directory = os.path.split(filename)[0]
        if os.path.exists(directory) and not os.listdir(directory):
            logger.warning("missing data %s", filename)
            return None

        raise InvalidParameter(f"directory={directory} does not exist.")

    return read_dataframe(filename, index_col=index_col, columns=columns)


def read_dataframe_by_substring(directory, substring, index_col=None, parse_dates=False, **kwargs):
    """Return a dataframe for the file containing substring.

    Parameters
    ----------
    directory : str
    substring : str
        identifier for output file, must be unique in directory
    index_col : str | int | None
        Index column name or index
    kwargs : kwargs
        Passed to underlying library for dataframe conversion.

    Returns
    -------
    pd.DataFrame

    """
    files = [x for x in os.listdir(directory) if substring in x]

    # Exclude any files that may have rolled, such as
    # Circuits-Losses-1-2.feather.1.gz
    regex = re.compile(r"\.\w+\.\d+(?:\.\w+)?$")
    files = [x for x in files if regex.search(x) is None]

    if not files:
        return None

    assert len(files) == 1, f"found multiple {substring} files in {directory}"

    filename = files[0]
    return read_dataframe(
        os.path.join(directory, filename), index_col=index_col, parse_dates=parse_dates, **kwargs
    )


def read_dataframes_by_substrings(
    directory, substrings, index_col=None, parse_dates=False, **kwargs
):
    """Return dataframes for files in the directory.

    Parameters
    ----------
    directory : str
    substrings : tuple
        File substring (str) for output files
    index_col : str | int | None
        Index column name or index
    kwargs : kwargs
        Passed to underlying library for dataframe conversion.

    Returns
    -------
    dict
        dictionary of substring to pd.DataFrame

    """
    dfs = {}

    for substring in substrings:
        df = read_dataframe_by_substring(
            directory,
            substring,
            index_col=index_col,
            parse_dates=parse_dates,
            **kwargs,
        )
        if df is not None:
            dfs[substring] = df

    return dfs


@timed_debug
def write_dataframe(df, file_path, compress=False, keep_original=False, **kwargs):
    """Write the dataframe to a file with in a format matching the extension.

    Note that the feather and h5 formats do not support row indices.
    Index columns will be lost for those formats. If the dataframe has an index
    then it should be converted to a column before calling this function.

    This function only supports storing a single dataframe inside an HDF5 file.
    It always uses the key 'data'.

    Parameters
    ----------
    df : pd.DataFrame
    file_path : str
    compress : bool
    keep_original : bool
    kwargs : pass keyword arguments to underlying library

    Raises
    ------
    InvalidParameter if the file extension is not supported.
    InvalidParameter if the DataFrame index is set.

    """
    if not isinstance(df.index, pd.RangeIndex) and not isinstance(
        df.index, pd.core.indexes.base.Index
    ):
        raise InvalidParameter("DataFrame index must not be set")

    ext = os.path.splitext(file_path)[1]

    if ext == ".csv":
        df.to_csv(file_path, **kwargs)
    elif ext == ".feather":
        df.to_feather(file_path, **kwargs)
    elif ext == ".h5":
        # HDF5 supports built-in compression, levels 1-9
        if "complevel" in kwargs:
            complevel = kwargs["complevel"]
        elif compress:
            complevel = 9
        else:
            complevel = 0
        df.to_hdf(file_path, "data", mode="w", complevel=complevel, **kwargs)
    elif ext == ".json":
        df.to_json(file_path, **kwargs)
    else:
        raise InvalidParameter(f"unsupported file extension {ext}")

    logger.debug("Created %s", file_path)

    if compress and ext != ".h5":
        zipped_path = file_path + ".gz"
        with open(file_path, "rb") as f_in:
            with gzip.open(zipped_path, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)

        if not keep_original:
            os.remove(file_path)

        file_path = zipped_path
        logger.debug("Compressed %s", zipped_path)

    return file_path


# TODO: broken with newer versions of pandas because of mishandling of indices
# def convert_csvs_to_feather(directory, compress=False, column_map=None,
#                            exclude_substrings=None, keep_original=False):
#    """Load CSV files as dataframes and convert them to feather files.
#
#    Parameters
#    ----------
#    directory : str
#        Convert all CSV files in the directory.
#    compress : bool
#        Compress the feather files.
#    column_map : dict
#        For CSV files without a header, pass basename-to-columns mapping
#    exclude_substrings : list
#        List of filename substrings to exclude.
#
#    """
#    for filename in os.listdir(directory):
#        skip = False
#        for substring in exclude_substrings:
#            if substring is not None and substring in filename:
#                skip = True
#                break
#        if skip:
#            continue
#        if os.path.splitext(filename)[1] == ".csv":
#            convert_csv_to_feather(os.path.join(directory, filename),
#                                   compress=compress,
#                                   column_map=column_map,
#                                   keep_original=keep_original)
#
#
# def convert_csv_to_feather(file_path, compress=False, column_map=None,
#                           keep_original=False):
#    """Load CSV files as dataframes and convert them to feather files.
#
#    Parameters
#    ----------
#    directory : str
#    filename : str
#    compress : bool
#        Compress the feather files.
#    column_map: dict
#        For CSV files without a header, pass basename-to-columns mapping
#
#    """
#    directory = os.path.dirname(file_path)
#    filename = os.path.basename(file_path)
#
#    assert os.path.splitext(filename)[1] == ".csv"
#
#    columns = None
#    if column_map is not None:
#        columns = column_map.get(os.path.basename(filename))
#
#    df = read_dataframe(os.path.join(directory, filename), columns=columns)
#    new_filename = os.path.splitext(filename)[0] + ".feather"
#    write_dataframe(df, os.path.join(directory, new_filename),
#                    compress=compress)
#
#    if not keep_original:
#        os.remove(os.path.join(directory, filename))
#
#    logger.debug("Converted %s: %s to %s", directory, filename, new_filename)



"""Common functions for CLI scripts"""

from jade.models.submitter_params import ResourceMonitorStats
import logging.config
import os
import sys

import click

from jade.common import HPC_CONFIG_FILE
from jade.enums import Mode, ResourceMonitorType
from jade.exceptions import UserAbort
from jade.models import (
    HpcConfig,
    LocalHpcConfig,
    SingularityParams,
    SubmitterParams,
    get_model_defaults,
)
from jade.utils.utils import load_data


logger = logging.getLogger(__name__)


_ENUM_MAPPING = {
    # class to type of value
    "mode": (Mode, str),
}


def handle_enum_input(_, param, value):
    """Converts inputs to enums."""
    try:
        tup = _ENUM_MAPPING.get(param.name)
        assert tup is not None, "Must add {} to _ENUM_MAPPING".format(param.name)
        # Force the correct type onto the value.
        return tup[0](tup[1](value))
    except ValueError as err:
        raise click.BadParameter(str(err))


def _handle_resource_monitor_type(_, __, value):
    if value is None:
        return None
    return ResourceMonitorType(value)


def _handle_simulation_scripts(ctx, _, value):
    if not os.path.exists(value):
        logger.info("Creating %s...", value)
        try:
            os.makedirs(value, exist_ok=True)
        except Exception as err:
            logger.error("Could not create %s: %s", value, err)
            raise
    else:
        proceed_with_user_permission(ctx, "Overwrite {}".format(value))
        try:
            scripts = [f for f in os.listdir(value) if f.endswith(".sh")]
            for s in scripts:
                os.remove(os.path.join(value, s))
        except Exception as err:
            logger.error("Could not delete %s: %s", value, err)
            raise

    return value


def proceed_with_user_permission(ctx, message):
    """Pauses execution to prompt user for permission to proceed."""
    prompt = ctx.params.get("prompt")
    if prompt is None:
        prompt = ctx.parent.params["prompt"]
    if not prompt:
        logger.info("Proceed because user passed --no-prompt.")
        return

    answer = ""
    while answer not in ("y", "n"):
        answer = input("{} [Y/N]? ".format(message)).lower()

    if answer == "n":
        raise UserAbort


SUBMITTER_PARAMS_DEFAULTS = get_model_defaults(SubmitterParams)

COMMON_SUBMITTER_OPTIONS = (
    click.option(
        "-C",
        "--container",
        type=click.Path(exists=True),
        help="Path to Singularity container",
    ),
    click.option(
        "-N",
        "--no-distributed-submitter",
        is_flag=True,
        default=False,
        show_default=True,
        help="Disable the distributed submitter",
    ),
    click.option(
        "-b",
        "--per-node-batch-size",
        default=SUBMITTER_PARAMS_DEFAULTS["per_node_batch_size"],
        show_default=True,
        help="Number of jobs to run on one node in one batch.",
    ),
    click.option(
        "--dry-run",
        default=False,
        is_flag=True,
        show_default=True,
        help="Dry run mode. Do not run any jobs. Not allowed in local mode.",
    ),
    click.option(
        "-h",
        "--hpc-config",
        type=click.Path(),
        default=HPC_CONFIG_FILE,
        show_default=True,
        help="HPC config file.",
    ),
    click.option(
        "-l",
        "--local",
        is_flag=True,
        default=False,
        show_default=True,
        envvar="LOCAL_SUBMITTER",
        help="Run on local system. Optionally, set the environment variable " "LOCAL_SUBMITTER=1.",
    ),
    click.option(
        "-n",
        "--max-nodes",
        default=SUBMITTER_PARAMS_DEFAULTS["max_nodes"],
        show_default=True,
        type=click.IntRange(
            2,
        ),
        help="Max number of node submission requests to make in parallel. Default is unbounded.",
    ),
    click.option(
        "-p",
        "--poll-interval",
        default=SUBMITTER_PARAMS_DEFAULTS["poll_interval"],
        type=float,
        show_default=True,
        help="Interval in seconds on which to poll jobs for status.",
    ),
    click.option(
        "-q",
        "--num-parallel-processes-per-node",
        "--num-processes",
        default=SUBMITTER_PARAMS_DEFAULTS["num_parallel_processes_per_node"],
        show_default=False,
        type=int,
        is_eager=True,
        help="Number of processes to run in parallel on each node; defaults to num CPUs.",
    ),
    click.option(
        "-m",
        "--resource-monitor-stats",
        multiple=True,
        type=click.Choice([x for x in ResourceMonitorStats.__fields__]),
        help="Resource stats to monitor. Default is CPU and memory. "
        "Ex: -m cpu -m memory -m process",
    ),
    click.option(
        "-r",
        "--resource-monitor-interval",
        default=None,
        type=int,
        show_default=True,
        help="Interval in seconds on which to collect resource stats. Must be >= poll-interval.",
    ),
    click.option(
        "-R",
        "--resource-monitor-type",
        default=None,
        type=click.Choice([x.value for x in ResourceMonitorType]),
        help="Type of resource monitoring to perform. Default is 'aggregation' unless "
        "--resource-monitor-interval is specified. In order to maintain backwards compatibility "
        "that changes the default to 'periodic'. "
        "'aggregation' will keep average/min/max stats in memory and generate a summary report "
        "at the end. 'periodic' will record stats to files at the specified interval. "
        "It will generate a summary report as well as interactive HTML plots.",
        callback=_handle_resource_monitor_type,
    ),
    click.option(
        "--reports/--no-reports",
        is_flag=True,
        default=SUBMITTER_PARAMS_DEFAULTS["generate_reports"],
        show_default=True,
        help="Generate reports after execution.",
    ),
    click.option(
        "-S",
        "--enable-singularity",
        is_flag=True,
        default=False,
        show_default=True,
        help="Add Singularity parameters and set the config to run in a container.",
    ),
    click.option(
        "-t",
        "--time-based-batching",
        is_flag=True,
        default=SUBMITTER_PARAMS_DEFAULTS["time_based_batching"],
        show_default=True,
        help="Use estimated runtimes to create batches. Each job must have its estimated runtime "
        "defined. Also requires --num-parallel-processes-per-node to be set. "
        "Overrides --per-node-batch-size.",
    ),
    click.option(
        "--try-add-blocked-jobs/--no-try-add-blocked-jobs",
        is_flag=True,
        default=SUBMITTER_PARAMS_DEFAULTS["try_add_blocked_jobs"],
        show_default=True,
        help="Add blocked jobs to a node's batch if they are blocked by jobs "
        "already in the batch.",
    ),
    click.option(
        "--verbose",
        is_flag=True,
        default=SUBMITTER_PARAMS_DEFAULTS["verbose"],
        show_default=True,
        help="Enable verbose log output.",
    ),
    click.option(
        "-x",
        "--node-setup-script",
        default=SUBMITTER_PARAMS_DEFAULTS["node_setup_script"],
        help="Deprecated. Script to run on each node before starting jobs (download input files).",
    ),
    click.option(
        "-y",
        "--node-shutdown-script",
        default=SUBMITTER_PARAMS_DEFAULTS["node_shutdown_script"],
        help="Deprecated. Script to run on each after completing jobs (upload output files).",
    ),
)


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def make_submitter_params(
    per_node_batch_size=None,
    dry_run=None,
    hpc_config=None,
    local=None,
    max_nodes=None,
    poll_interval=None,
    resource_monitor_interval=None,
    resource_monitor_type=None,
    resource_monitor_stats=None,
    num_parallel_processes_per_node=None,
    verbose=None,
    reports=None,
    enable_singularity=None,
    container=None,
    try_add_blocked_jobs=None,
    time_based_batching=None,
    node_setup_script=None,
    node_shutdown_script=None,
    no_distributed_submitter=None,
):
    """Returns an instance of SubmitterParams for use in a job submission."""
    if node_setup_script is not None or node_shutdown_script is not None:
        print(
            "Warning: node_setup_script and node_shutdown_script are deprecated and will "
            "be removed in release v0.9.0."
        )
    if local:
        hpc_config = HpcConfig(hpc_type="local", hpc=LocalHpcConfig())
    else:
        # TODO: If the config file contains submission groups then this should not be required.
        if not os.path.exists(hpc_config):
            print(
                f"{hpc_config} does not exist. Generate it with 'jade config hpc' "
                "or run in local mode with '-l'",
                file=sys.stderr,
            )
            sys.exit(1)
        hpc_config = HpcConfig(**load_data(hpc_config))

    if local and dry_run:
        print("Dry run is not allowed in local mode.", file=sys.stderr)
        sys.exit(1)

    if (
        time_based_batching
        and per_node_batch_size != SUBMITTER_PARAMS_DEFAULTS["per_node_batch_size"]
    ):
        # This doesn't catch the case where the user passes --per-node-batch-size=default, but
        # I don't see that click provides a way to detect that condition.
        print(
            "Error: --per-node-batch-size and --time-based-batching are mutually exclusive",
            file=sys.stderr,
        )
        sys.exit(1)

    if time_based_batching and num_parallel_processes_per_node is None:
        print(
            "Error: num_parallel_processes_per_node must be set with time-based batching",
            file=sys.stderr,
        )
        sys.exit(1)

    # We added resource_monitor_type after resource_monitor_interval. The following logic
    # maintains backwards compatibility with user settings.
    default_monitor_interval = SUBMITTER_PARAMS_DEFAULTS["resource_monitor_interval"]
    if resource_monitor_interval is not None and resource_monitor_type is not None:
        pass
    elif resource_monitor_interval is None and resource_monitor_type is None:
        resource_monitor_type = ResourceMonitorType.AGGREGATION
        resource_monitor_interval = default_monitor_interval
    elif resource_monitor_interval is not None and resource_monitor_type is None:
        resource_monitor_type = ResourceMonitorType.PERIODIC
    elif resource_monitor_interval is None and resource_monitor_type is not None:
        resource_monitor_interval = default_monitor_interval
    else:
        assert False, f"interval={resource_monitor_interval} type={resource_monitor_type}"
    if not resource_monitor_stats:
        resource_monitor_stats = ResourceMonitorStats()
    else:
        stats = {x: True for x in resource_monitor_stats}
        for field in ResourceMonitorStats.__fields__:
            if field not in stats:
                stats[field] = False
        resource_monitor_stats = ResourceMonitorStats(**stats)

    if resource_monitor_interval is not None:
        if resource_monitor_interval < poll_interval:
            logger.warning(
                "resource_monitor_interval cannot be less than poll_interval. "
                "Reducing poll_interval"
            )
            poll_interval = resource_monitor_interval

    if enable_singularity:
        singularity_params = SingularityParams(enabled=True, container=container)
    else:
        singularity_params = None
    return SubmitterParams(
        generate_reports=reports,
        hpc_config=hpc_config,
        max_nodes=max_nodes,
        num_parallel_processes_per_node=num_parallel_processes_per_node,
        per_node_batch_size=per_node_batch_size,
        distributed_submitter=not no_distributed_submitter,
        dry_run=dry_run,
        node_setup_script=node_setup_script,
        node_shutdown_script=node_shutdown_script,
        poll_interval=poll_interval,
        resource_monitor_interval=resource_monitor_interval,
        resource_monitor_type=resource_monitor_type,
        resource_monitor_stats=resource_monitor_stats,
        singularity_params=singularity_params,
        time_based_batching=time_based_batching,
        try_add_blocked_jobs=try_add_blocked_jobs,
        verbose=verbose,
    )



"""
Implement the JobParametersInterface for auto-regression analysis.
"""

from collections import namedtuple

from jade.common import DEFAULT_SUBMISSION_GROUP
from jade.jobs.job_parameters_interface import JobParametersInterface


class AutoRegressionParameters(JobParametersInterface):
    """
    A class used for creating auto-regression job.
    """

    parameters_type = namedtuple("AutoRegression", "country")
    _EXTENSION = "demo"

    def __init__(self, country, data):
        """
        Init auto-regression parameter class

        Parameters
        ----------
        country: str
            The name of a country.
        data: str
            The path to the csv file containing the GDP data.
        """
        self.country = country
        self.data = data
        self._name = self._create_name()

    def __str__(self):
        return "<AutoRegressionParameters: {}>".format(self.name)

    @property
    def extension(self):
        return self._EXTENSION

    @property
    def name(self):
        return self._name

    def _create_name(self):
        return self.country.replace(" ", "_").lower()

    def serialize(self):
        return {
            "country": self._name,
            "data": self.data,
            "extension": self.extension,
        }

    @classmethod
    def deserialize(cls, param):
        return cls(param["country"], param["data"])

    @property
    def estimated_run_minutes(self):
        return None

    @property
    def cancel_on_blocking_job_failure(self):
        return False

    def get_blocking_jobs(self):
        return set()

    def remove_blocking_job(self, name):
        pass

    def set_blocking_jobs(self, blocking_jobs):
        pass

    @property
    def submission_group(self):
        return DEFAULT_SUBMISSION_GROUP

    @submission_group.setter
    def submission_group(self, group):
        return



"""
setup.py
"""
import os
import logging
from codecs import open
from pathlib import Path
from setuptools import setup, find_packages


logger = logging.getLogger(__name__)


def read_lines(filename):
    return Path(filename).read_text().splitlines()


here = os.path.abspath(os.path.dirname(__file__))

with open("README.md", encoding="utf-8") as f:
    readme = f.read()

with open(os.path.join(here, "jade", "version.py"), encoding="utf-8") as f:
    version = f.read()

version = version.split()[2].strip('"').strip("'")

# TODO: Current code breaks with statsmodels 0.13.
demo_requires = ["matplotlib", "statsmodels"]
dataframe_utils_requires = ["tables"]
dev_requires = read_lines("dev-requirements.txt") + demo_requires + dataframe_utils_requires


setup(
    name="NREL-jade",
    version=version,
    description="Provides HPC workflow automation services",
    long_description=readme,
    long_description_content_type="text/markdown",
    maintainer="Daniel Thom",
    maintainer_email="daniel.thom@nrel.gov",
    url="https://github.com./NREL/jade",
    packages=find_packages(),
    package_dir={"jade": "jade"},
    entry_points={
        "console_scripts": [
            "jade=jade.cli.jade:cli",
            "jade-internal=jade.cli.jade_internal:cli",
        ],
    },
    include_package_data=True,
    license="BSD license",
    zip_safe=False,
    keywords=["jade", "hpc", "workflow"],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
    ],
    test_suite="tests",
    extras_require={
        "dev": dev_requires,
        "demo": demo_requires,
        "dataframe_utils": dataframe_utils_requires,
    },
    install_requires=read_lines("requirements.txt"),
)
